﻿using AsReaderP252B;
using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.IO;
using System.IO.Ports;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;
using System.Timers;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Media.Imaging;
using System.Configuration;

namespace AsReaderP252B_Demo
{
    /// <summary>
    /// Main.xaml 的交互逻辑
    /// </summary>
    public partial class Main : Page
    {
        ConnectStatus connectStatus = ConnectStatus.Disconnected;
        int _batteryInterval = 10;
        int _defaultLinkProfile = 1;
        int SettingBatteryInterval = 0;
        int SettingDefaultLinkProfile = 0;
        BluetoothDevice SettingDefaultBTDevice;
        System.Timers.Timer ConnectBTDeviceTimer;
        System.Timers.Timer GetBatteryTimer;
        System.Timers.Timer STimerInventory;
        System.Timers.Timer RescanBarcodeTimer;
        System.Threading.Timer TRescanBarcodeTimer;
        System.Threading.Timer InventoryTimer;
        //System.Timers.Timer GetTemperatureTimer;
        ActionType actionType = ActionType.None;
        string[] freqArr = { };
        string curFrequency = "";
        DeviceSearch deviceSearchUI;
        BluetoothDevice currentBluetoothDevice;
        public List<Bind> KVFrequencyList { get; set; }
        ObservableCollection<Tag> TagList = new ObservableCollection<Tag>();
        ObservableCollection<Barcode> BarcodeList = new ObservableCollection<Barcode>();
        ConnectTypes _connectTypes = ConnectTypes.None;
        private static AutoResetEvent resetEvent = new AutoResetEvent(false);
        bool _isBTDisConnClick = false;
        int? rcvBattery = null;
        public Main()
        {
            InitializeComponent();
            SettingDefaultBTDevice = Properties.Settings.Default.BluetoothDevice;
            SetDelegateMethod();
            Device.SetDelegates();
            EnableControllers();
            GetComPorts();
            SetComboxList();
            Dispatcher.ShutdownStarted += ShutdownStarted;
            int.TryParse(Properties.Settings.Default.BatteryInterval, out SettingBatteryInterval);
            int.TryParse(Properties.Settings.Default.DefaultLinkProfile, out SettingDefaultLinkProfile);
            InitTimers();
            if (SettingDefaultBTDevice != null) ConnectBTDeviceTimer.Start();
            lblWarning.Visibility = Visibility.Hidden;
        }
        private void ShutdownStarted(object sender, EventArgs e)
        {
            if (connectStatus == ConnectStatus.Connected) Device.Stop();
            Properties.Settings.Default.Save();
            Configuration cfa = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None);
            cfa.AppSettings.Settings["AdminPwd"].Value = "";
            cfa.Save();
        }
        void InitTimers()
        {
            RescanBarcodeTimer = new System.Timers.Timer();
            RescanBarcodeTimer.Elapsed += new System.Timers.ElapsedEventHandler(RescanBarcode);

            STimerInventory = new System.Timers.Timer();
            STimerInventory.Elapsed += new System.Timers.ElapsedEventHandler(InventoryTimerDo);
            STimerInventory.Interval = 1000;

            ConnectBTDeviceTimer = new System.Timers.Timer();
            ConnectBTDeviceTimer.Elapsed += new System.Timers.ElapsedEventHandler(ConnectBTDeviceAuto);
            ConnectBTDeviceTimer.Interval = 7000;

            //GetTemperatureTimer = new System.Timers.Timer();
            //GetTemperatureTimer.Interval = 5000;
            //GetTemperatureTimer.Elapsed += new System.Timers.ElapsedEventHandler(GetTemperature);
            //GetTemperatureTimer.AutoReset = true;
        }
        void GetTemperature(object sender, ElapsedEventArgs e)
        {
            Device.GetTemperature();
        }
        void RescanBarcode(object sender, ElapsedEventArgs e)
        {
            Device.StartBarcodeScan();
        }
        int inventorySecondCount = 0;
        void ConnectBTDeviceAuto(object sender, ElapsedEventArgs e)
        {
            if (connectStatus == ConnectStatus.Connecting || connectStatus == ConnectStatus.Connected) return;
            SettingDefaultBTDevice = Properties.Settings.Default.BluetoothDevice;
            if (SettingDefaultBTDevice == null)
            {
                ConnectBTDeviceTimer.Stop();
            }
            else
            {
                Dispatcher.BeginInvoke((Action)(() =>
                {
                    currentBluetoothDevice = SettingDefaultBTDevice;
                    lblBTName.Content = currentBluetoothDevice.DeviceName;
                    btnBTConnect_Click(null, null);
                }));
            }
        }
        int[] countArr = new int[5];
        int arrTop = -1;
        void InventoryTimerDo(object sender, ElapsedEventArgs e)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                inventorySecondCount++;
                lblTime.Content = inventorySecondCount.ToString();
                //计算平均速度
                //int diff = readTagCount - readTagCountBK;
                //if (arrTop >= 4) arrTop = -1;
                //countArr[++arrTop] = diff;
                //int total = 0;
                //string tmp = "";
                //for (int i = 0; i < countArr.Length; i++)
                //{
                //    tmp += countArr[i] + " ";
                //    total += countArr[i];
                //}
                //Console.WriteLine($"<<{readTagCount} - {readTagCountBK} = {diff}");
                //Console.WriteLine(">>" + tmp);
                //if (inventorySecondCount >= 5) lblSpeed.Content = total / 5;//5秒内平均速度
                //else lblSpeed.Content = total / inventorySecondCount;
                //当前秒的总盘点数量-前一秒的盘点数量=盘点速度
                lblSpeed.Content = readTagCount - readTagCountBK;
                readTagCountBK = readTagCount;
            }));
        }
        void SetBatteryInterverTimer(int interval)
        {
            if (GetBatteryTimer == null)
            {
                GetBatteryTimer = new System.Timers.Timer();
                GetBatteryTimer.Elapsed += new System.Timers.ElapsedEventHandler(GetBattery);
                GetBatteryTimer.AutoReset = true;
            }
            GetBatteryTimer.Interval = interval * 1000;
        }
        void GetBattery(object sender, ElapsedEventArgs e)
        {
            Device.GetBattery();
        }
        void GetComPorts()
        {
            string[] ports = SerialPort.GetPortNames();
            cmbPorts.Items.Clear();
            Console.WriteLine($"COM ports count: {ports.Length}");
            for (int i = 0; i < ports.Length; i++)
            {
                cmbPorts.Items.Add(ports[i]);
            }
            if (cmbPorts.Items.Count > 0) cmbPorts.SelectedIndex = 0;
        }
        private void SetComboxList()
        {
            DataTable dtTmp = new DataTable();
            dtTmp = Device.getComboxList(typeof(Types.Baudrate));
            cmbBaundrate.ItemsSource = dtTmp.DefaultView;
            cmbBaundrate.DisplayMemberPath = "name";
            cmbBaundrate.SelectedIndex = 2;

            //cmbOptionBaudRate.ItemsSource = dtTmp.Copy().DefaultView;
            //cmbOptionBaudRate.DisplayMemberPath = "name";

            dtTmp = Device.getComboxList(typeof(Types.BuzzerStatus));
            cmbBuzzer.ItemsSource = dtTmp.DefaultView;
            cmbBuzzer.DisplayMemberPath = "name";

            string[] list = Device.GetQValueList();
            cmbQValue.ItemsSource = list;
            cmbMaxQ.ItemsSource = list;
            cmbMinQ.ItemsSource = list;

            string[] sessionList = { "S0", "S1", "S2", "S3" };
            cmbIvtSession.ItemsSource = sessionList;

            string[] sessionFlagList = { "A Only", "B Only", "A or B" };
            cmbSessionFlag.ItemsSource = sessionFlagList;

            int[] arr = new int[4] { 0, 1, 2, 3 };
            cmbLinkProfile.ItemsSource = arr;
            cmbDefaultLinkProfile.ItemsSource = arr;

            Dictionary<int, string> dicMaskList = new Dictionary<int, string>();
            for (int i = 0; i <= 30; i += 2)
            {
                dicMaskList.Add(8 * i, (8 * i).ToString() + " bit");
            }
            //cmbMaskOffset.ItemsSource = maskList;
            //cmbMaskLength.ItemsSource = maskList;
            cmbMaskOffset.DataContext = dicMaskList;
            cmbMaskOffset.SetBinding(ComboBox.ItemsSourceProperty, new Binding());
            cmbMaskLength.DataContext = dicMaskList;
            cmbMaskLength.SetBinding(ComboBox.ItemsSourceProperty, new Binding());

            List<string> rescanList = new List<string>();
            int baseData = 0;
            for (int i = 0; i < 20; i++)
            {
                if (i == 0) rescanList.Add("Not Used");
                else if (i <= 13)
                {
                    baseData = baseData + 100;
                    rescanList.Add(baseData.ToString() + " ms");
                }
                else if (i == 14)
                {
                    baseData = baseData + 200;
                    rescanList.Add(baseData.ToString() + " ms");
                }
                else if (i > 14 && i < 19)
                {
                    baseData = baseData + 500;
                    rescanList.Add(baseData.ToString() + " ms");
                }
                else rescanList.Add("5000 ms");
            }
            cmbRescanTime.ItemsSource = rescanList;
            cmbRescanTime.SelectedIndex = 0;

            List<string> wordList = new List<string>();
            for (int i = 0; i < 16; i++)
            {
                wordList.Add(i.ToString() + " WORD");
            }
            cmbReadOffset.ItemsSource = wordList;
            cmbReadLength.ItemsSource = wordList;
            cmbWriteOffset.ItemsSource = wordList;
            cmbReadOffset.SelectedIndex = cmbReadLength.SelectedIndex = cmbWriteOffset.SelectedIndex = 2;

            string[] onOffList = { "Off", "On" };
            cmbTemperatureOnOff.ItemsSource = onOffList;
            string[] overrunAction = { "Only Warning", "Warning and interval reading" , "Warning and stop reading" };
            cmbOverrunAction.ItemsSource = overrunAction;
        }
        void ClearInformation()
        {
            lblFirmwareVer.Content = "";
            lblHardVer.Content = "";
            lblRegionCoutry.Content = "";
            lblSN.Content = "";
            lblUhfHf.Content = "";
        }
        void ClearOption()
        {
            cmbBuzzer.SelectedIndex = -1;
            txtIdle.Text = "";
            txtSleepTime.Text = "";
            txtBatteryInterval.Text = "";
            txtAutoOffTime.Text = "";
            cmbQValue.SelectedIndex = -1;
            cmbMaxQ.SelectedIndex = -1;
            cmbMinQ.SelectedIndex = -1;
            cmbLinkProfile.SelectedIndex = -1;
            cmbDefaultLinkProfile.SelectedIndex = -1;
            txtBarcodeTimeout.Text = "";
            cmbOptionBaudRate.SelectedIndex = -1;
            cbFrequencyAuto.IsChecked = false;
            lblLBTChannel.Visibility = Visibility.Hidden;
            lvBTChannel.Visibility = Visibility.Hidden;
        }
        private void DataGrid_MouseDown(object sender, MouseButtonEventArgs e)
        {
            if (e.LeftButton == MouseButtonState.Pressed)
            {
                var se = dgBarcodes.SelectedItem;
                FrameworkElement objElement = dgBarcodes.Columns[0].GetCellContent(se);
                if (objElement != null)
                {
                    CheckBox objChk = (CheckBox)objElement;
                    objChk.IsChecked = !objChk.IsChecked;
                }
            }
        }
        private void lblOperationHeader_MouseLeftButtonDown(object sender, RoutedEventArgs e)
        {
            if (connectStatus == ConnectStatus.Connected)
            {
                Device.GetPower();
                Thread.Sleep(50);
                Device.GetOperationTime();
                Thread.Sleep(50);
                Device.GetRssiOnOff();
            }
        }
        private void DataGridRow_GotFocus(object sender, RoutedEventArgs e)
        {
            DataGridRow item = (DataGridRow)sender;
            string epc = ((TextBlock)dgTags.Columns[4].GetCellContent(item)).Text;
            //var item = (DataGridRow)sender;
            //FrameworkElement objElement = dgTags.Columns[0].GetCellContent(item);
            //if (objElement != null)
            //{
            //    CheckBox objChk = (CheckBox)objElement;
            //    objChk.IsChecked = !objChk.IsChecked;
            //}
        }
        void EnableControllers()
        {
            bool isConnected = connectStatus == ConnectStatus.Connected ? true : false;
            tbLeft.IsEnabled = isConnected;
            tbRight.IsEnabled = isConnected;
            btnDisconnect.IsEnabled = isConnected && _connectTypes == ConnectTypes.COM;
            btnConnect.IsEnabled = !isConnected;
            btnBTDisconnect.IsEnabled = isConnected && _connectTypes == ConnectTypes.Bluetooth;
            btnBTConnect.IsEnabled = !isConnected;
            imgBattery.IsEnabled = isConnected;
            btnSearch.IsEnabled = !isConnected;
            Visibility tbItemTempVisibility = Visibility.Hidden;
            string adminPwd = ConfigurationManager.AppSettings["AdminPwd"];
            if (!string.IsNullOrEmpty(adminPwd))
            {
                MD5 md5 = new MD5CryptoServiceProvider();
                byte[] result = md5.ComputeHash(System.Text.Encoding.Default.GetBytes(adminPwd));
                string text = BitConverter.ToString(result).Replace("-", "");
                if (Properties.Settings.Default.AdminPwdKey == text.ToLower())
                {
                    tbItemTempVisibility = Visibility.Visible;
                }
            }
            tbItemTemp.Visibility = tbItemTempVisibility;
        }
        void ShowMessage(string message)
        {
            txtLog.AppendText(DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss") + " " + message + Environment.NewLine);
            txtLog.ScrollToEnd();
        }
        void SetDelegateMethod()
        {
            Device.whenReceivedFirmwareVersion = ReceivedFirmwareVersion;
            Device.whenReceivedHardVersion = ReceivedHardVersion;
            Device.whenReceivedRegionCountry = ReceivedRegionCountry;
            Device.whenReceivedSerialNumber = ReceivedSerialNumber;
            Device.whenReceivedRFIDModuleVersion = ReceivedRFIDModuleVersion;
            Device.whenReceivedBuzzer = ReceivedBuzzer;
            Device.whenReceivedAutoOffSleep = ReceivedAutoOffSleep;
            Device.whenReceivedAutoOffTime = ReceivedAutoOffTime;
            Device.whenReceivedQValue = ReceivedQValue;
            Device.whenReceivedQMaxValue = ReceivedQMaxValue;
            Device.whenReceivedQMinValue = ReceivedQMinValue;
            Device.whenReceivedIdleTime = ReceivedIdleTime;
            Device.whenReceivedBattery = ReceivedBattery;
            Device.whenReceivedLinkProfile = ReceivedLinkProfile;
            Device.whenReceivedBaudRate = ReceivedBaudRate;
            Device.whenReceivedBaudRateList = ReceivedBaudRateList;
            Device.whenReceivedBarcodeTimeout = ReceivedBarcodeTimeout;
            Device.whenReceivedFrequencyList = ReceivedFrequencyList;
            Device.whenReceivedFrequency = ReceivedFrequency;
            Device.whenReceivedRestoreFactorySettings = ReceivedRestoreFactorySettings;
            Device.whenReceivedSetBuzzerResult = ReceivedSetBuzzerResult;
            Device.whenReceivedSetIdleTimeResult = ReceivedSetIdleTimeResult;
            Device.whenReceivedSetAutoOffSleepResult = ReceivedSetAutoOffSleepResult;
            Device.whenReceivedSetFrequencyResult = ReceivedSetFrequencyResult;
            Device.whenReceivedSetAutoOffTimeResult = ReceivedSetAutoOffTimeResult;
            Device.whenReceivedSetBarcodeTimeoutResult = ReceivedSetBarcodeTimeoutResult;
            Device.whenReceivedStartBarcodeScanResult = ReceivedStartBarcodeScanResult;
            Device.whenReceivedStopResult = ReceivedStopResult;
            Device.whenReceivedBarcodeData = ReceivedBarcodeData;
            Device.whenReceivedEPCMaskData = ReceivedEPCMaskData;
            Device.whenReceivedClearEPCMaskResult = ReceivedClearEPCMaskResult;
            Device.whenReceivedSetEPCMaskResult = ReceivedSetEPCMaskResult;
            Device.whenReceivedSetPasswordResult = ReceivedSetPasswordResult;
            Device.whenReceivedReadResult = ReceivedReadResult;
            Device.whenReceivedWriteResult = ReceivedWriteResult;
            Device.whenReceivedLockResult = ReceivedLockResult;
            Device.whenReceivedKillResult = ReceivedKillResult;
            Device.whenReceivedOperationTime = ReceivedOperationTime;
            Device.whenReceivedInventoryMode = ReceivedInventoryMode;
            Device.whenReceivedRssiOnOff = ReceivedRssiOnOff;
            Device.whenReceivedSessionFlag = ReceivedSessionFlag;
            Device.whenReceivedSetSessionResult = ReceivedSetSessionResult;
            Device.whenReceivedSetRssiOnOffResult = ReceivedSetRssiOnOffResult;
            Device.whenReceivedSetSessionFlagResult = ReceivedSetSessionFlagResult;
            Device.whenReceivedSetInventoryModeResult = ReceivedSetInventoryModeResult;
            Device.whenReceivedSetPowerGainResult = ReceivedSetPowerGainResult;
            Device.whenReceivedSetOperationTimeResult = ReceivedSetOperationTimeResult;
            Device.whenReceivedPowerGain = ReceivedPowerGain;
            Device.whenReceivedSession = ReceivedSession;
            Device.whenReceivedStartInventoryResult = ReceivedStartInventoryResult;
            Device.whenReceivedTagData = ReceivedTagData;
            Device.whenReceivedPowerRange = ReceivedPowerRange;
            Device.whenReceivedReadTagData = ReceivedReadTagData;
            Device.whenReceivedTriggerStatus = ReceivedTriggerStatus;
            Device.whenReceivedModeStatus = ReceivedModeStatus;
            Device.whenReceivedSetQValueResult = ReceivedSetQValueResult;
            Device.whenReceivedSetLinkProfileResult = ReceivedSetLinkProfileResult;
            Device.whenReceivedSetBaudRateResult = ReceivedSetBaudRateResult;
            Device.whenReceivedPermaLockResult = ReceivedPermaLockResult;
            Device.whenDeviceIsConnected = DeviceIsConnected;
            Device.whenReceivedSetTemperatureResult = ReceivedSetTemperatureResult;
            Device.whenReceivedTemperature = ReceivedTemperature;
            Device.whenReceivedWorkTemperature = ReceivedWorkTemperature;


        }
        #region delegate functions
        void ReceivedWorkTemperature(string workTemp, string threshold)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                //ShowMessage("Received Work Temperature.");
                //lblWorkTemperature.Content = workTemp + " ℃";
                //lblWarning.Visibility = Visibility.Visible;
            }));
        }
        void ReceivedTemperature(string currentTemp)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Received Temperature.");
                //cmbTemperatureOnOff.SelectedIndex = onOff;
                //txtThreshold.Text = threshold;
                //cmbOverrunAction.SelectedIndex = action;
                lblCurrentTemp.Content = currentTemp + " ℃";
                lblWorkTemperature.Content = currentTemp + " ℃";
            }));
        }
        void ReceivedSetTemperatureResult()
        {
            Dispatcher.BeginInvoke(new Action(() =>
            {
                ShowMessage("Set Temperature succeed.");
            }));
        }
        void DeviceIsConnected(bool isConnected)
        {
            Dispatcher.BeginInvoke(new Action(() =>
            {
                ShowMessage($"DeviceIsConnected: {isConnected}");
            }));
            if (connectStatus != ConnectStatus.Connected && isConnected)
            {
                if (_connectTypes == ConnectTypes.Bluetooth)
                {
                    Thread.Sleep(1000);
                    Device.GetBattery();
                    resetEvent.WaitOne(2000);
                    if (!rcvBattery.HasValue) return;
                    else
                    {
                        if (ConnectBTDeviceTimer.Enabled)
                        {
                            ConnectBTDeviceTimer.Stop();
                        }
                    }
                }
            }
            this.Dispatcher.Invoke(new Action(() =>
            {
                if (!isConnected && btnBTDisconnect.IsEnabled)
                {
                    rcvBattery = null;
                    if (tbRight.SelectedIndex == 0 && btnInventory.Content.ToString() == "Stop") btnStartInventory_Click(null, null);
                    if (tbRight.SelectedIndex == 1 && btnScanBarcode.Content.ToString() == "Stop Scan") btnScanBarcode_Click(null, null);
                    WhenDisconnected();
                }
                else
                {
                    _connectTypes = ConnectTypes.Bluetooth;
                    WhenConnected(isConnected);
                }
            }));
        }
        void ReceivedPermaLockResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Permalock succeed.");
            }));
        }
        void ReceivedSetBaudRateResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set Baud Rate succeed.");
                if (!string.IsNullOrEmpty(SetBaudRateValue))
                {
                    Properties.Settings.Default.BaudRate = SetBaudRateValue;
                    Properties.Settings.Default.Save();
                }
            }));
        }
        void ReceivedSetLinkProfileResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set Link Profile succeed.");
                if (!string.IsNullOrEmpty(SetLinkProfileValue))
                {
                    Properties.Settings.Default.LinkProfile = SetLinkProfileValue;
                    Properties.Settings.Default.Save();
                }
            }));
        }
        void ReceivedSetQValueResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set QValue succeed.");
                if (!string.IsNullOrEmpty(SetQValueValue))
                {
                    Properties.Settings.Default.QValue = SetQValueValue;
                    Properties.Settings.Default.Save();
                }
            }));
        }
        void ReceivedModeStatus(Types.ButtonActionStatus buttonActionStatus)
        {
            string info = "Received mode status: " + Enum.GetName(typeof(Types.ButtonActionStatus), buttonActionStatus);
            Dispatcher.BeginInvoke((Action)(() =>
            {
                ShowMessage(info);
            }));
            //Log.Info(info);
        }
        void ReceivedTriggerStatus(Types.ButtonActionStatus buttonActionStatus)
        {
            string info = "Received trigger status: " + Enum.GetName(typeof(Types.ButtonActionStatus), buttonActionStatus);
            Dispatcher.BeginInvoke((Action)(() =>
            {
                ShowMessage(info);
                if (buttonActionStatus == Types.ButtonActionStatus.Down)
                {
                    if (tbRight.SelectedIndex == 1) btnScanBarcode_Click(null, null);
                    else if (tbRight.SelectedIndex == 0) btnStartInventory_Click(null, null);
                }
            }));
            //Log.Info(info);
        }
        void ReceivedReadTagData(Types.CommandType commandType, Types.ResultCode resultCode, TagData tagData, string[] readMemory)
        {
            Dispatcher.Invoke((Action)delegate ()
            {
                if (btnLED.Content.ToString() == "LED OFF")
                {
                    for (int i = 0; i < TagList.Count; i++)
                    {
                        if (TagList[i].TagData.EPC == tagData.EPC && TagList[i].TagData.PC == tagData.PC)
                        {
                            TagList[i].Count++;
                            TagList[0].TotalCount++;
                            TagList[i].TagData = tagData;
                            dgTags.ItemsSource = null;
                            dgTags.ItemsSource = TagList;
                            dgTags.SelectedItem = dgTags.Items[i];
                            lblReadCount.Content = TagList[0].TotalCount;
                            break;
                        }
                    }
                }
                Console.WriteLine(DateTime.Now.ToString("G") + "Read: " + tagData.EPC);
                string resultCodeName = Enum.GetName(typeof(Types.ResultCode), resultCode);
                string commandTypeName = Enum.GetName(typeof(Types.CommandType), commandType);
                switch (commandType)
                {
                    case Types.CommandType.ReadMemory:
                        txtReadData.Text = tagData.EPC;
                        lblReadRssi.Content = cbReadReportRssi.IsChecked == true ? tagData.Rssi.ToString("F1") + " dB" : "";
                        lblReadPhase.Content = cbReadReportRssi.IsChecked == true ? tagData.Phase.ToString("F0") + " °" : "";
                        int readWordCount = readMemory.Length;
                        if (readMemory != null && readWordCount > 0)
                        {
                            string readMem = "";
                            for (int i = 0; i < readWordCount; i++)
                            {
                                if (i < 8) readMem += (i == 0 ? "" : " ") + readMemory[i] + ((i == 7 && i < readWordCount - 1) ? System.Environment.NewLine + System.Environment.NewLine : "");
                                else if (i < 16) readMem += (i == 8 ? "" : " ") + readMemory[i];
                            }
                            txtReadMemory.Text = readMem;
                        }
                        break;
                    case Types.CommandType.WriteMemory:
                        if (tbLeft.SelectedIndex == 4)
                        {
                            txtWriteReadData.Text = tagData.EPC;
                            lblWriteRssi.Content = cbWriteReportRssi.IsChecked == true ? tagData.Rssi.ToString("F1") + " dB" : "";
                            lblWritePhase.Content = cbWriteReportRssi.IsChecked == true ? tagData.Phase.ToString("F0") + " °" : "";
                        }
                        else if (tbLeft.SelectedIndex == 5)
                        {
                            txtLockReadData.Text = tagData.EPC;
                            lblLockRssi.Content = cbLockReportRssi.IsChecked == true ? tagData.Rssi.ToString("F1") + " dB" : "";
                            lblLockPhase.Content = cbLockReportRssi.IsChecked == true ? tagData.Phase.ToString("F0") + " °" : "";
                        }
                        break;
                    case Types.CommandType.PermaLock:
                    case Types.CommandType.Kill:
                    case Types.CommandType.Lock:
                        txtLockReadData.Text = tagData.EPC;
                        lblLockRssi.Content = cbLockReportRssi.IsChecked == true ? tagData.Rssi.ToString("F1") + " dB" : "";
                        lblLockPhase.Content = cbLockReportRssi.IsChecked == true ? tagData.Phase.ToString("F0") + " °" : "";
                        break;
                }
                if (resultCode != Types.ResultCode.NoError)
                {
                    ShowMessage($"{commandTypeName}: {resultCodeName}");
                }
                SwitchButtonContent();
            });
        }
        void ReceivedPowerRange(int min, int max)
        {
            Dispatcher.Invoke((Action)delegate ()
            {
                List<string> powerList = new List<string>();
                for (int i = min; i <= max; i += 10)
                {
                    powerList.Add((i / 10f).ToString("#0.0") + " dB");
                }
                cmbPowerGain.ItemsSource = powerList;
                cmbReadPowerGain.ItemsSource = powerList;
                cmbWritePowerGain.ItemsSource = powerList;
                cmbLockPowerGain.ItemsSource = powerList;
            });
        }
        int readTagCount = 0;
        int readTagCountBK = 0;
        void ReceivedTagData(TagData tagData)
        {
            try
            {
                Task.Run(() =>
                {
                    //Dispatcher.BeginInvoke(System.Windows.Threading.DispatcherPriority.Normal, (Action)delegate ()
                    //{
                    lock (TagList)
                    {
                        readTagCount++;
                        bool isUpdate = false;
                        for (int i = 0; i < TagList.Count; i++)
                        {
                            if (TagList[i].TagData.PC == tagData.PC && TagList[i].TagData.EPC == tagData.EPC)
                            {
                                isUpdate = true;
                                TagList[i].Count += 1;
                                TagList[i].TagData = tagData;
                                TagList[0].TotalCount = readTagCount;
                                break;
                            }
                        }
                        if (!isUpdate)
                        {
                            Tag tag = new Tag(TagList.Count + 1, tagData, 1, 0);
                            TagList.Add(tag);
                            TagList[0].TotalCount = readTagCount;
                        }
                    }
                });
            }
            catch (Exception ex)
            {
                Console.WriteLine($"ReceivedTagData: {ex.Message}");
            }
        }
        void ShowTags(object state)
        {
            Dispatcher.BeginInvoke(System.Windows.Threading.DispatcherPriority.Background, (Action)(() =>
            {
                ObservableCollection<Tag> tmp;
                lock (TagList)
                {
                    tmp = new ObservableCollection<Tag>(TagList);
                }
                dgTags.ItemsSource = null;
                dgTags.ItemsSource = tmp;
                lblTagCount.Content = tmp.Count;
                lblReadCount.Content = tmp.Count > 0 ? tmp[0].TotalCount : 0;
            }));
        }
        void ReceivedStartInventoryResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Start inventory succeed.");
            }));
        }
        void ReceivedSession(int session)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                ShowMessage("Received session.");
                cmbIvtSession.SelectedIndex = session;
            }));
            //Log.Info($"Received session: {session}");
        }
        int ReceivedPowerGainIndex = -1;
        void ReceivedPowerGain(int power)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                ShowMessage("Received power gain." + power);
                int index = cmbPowerGain.Items.IndexOf((power / 10f).ToString("#0.0") + " dB");
                ReceivedPowerGainIndex = index;
                cmbPowerGain.SelectedIndex = index;
                cmbReadPowerGain.SelectedIndex = index;
                cmbWritePowerGain.SelectedIndex = index;
                cmbLockPowerGain.SelectedIndex = index;
            }));
            //Log.Info($"Received power gain: {power}");
        }
        void ReceivedSetOperationTimeResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set operation time succeed.");
                if (!string.IsNullOrEmpty(SetOperationTimeValue))
                {
                    Properties.Settings.Default.OperationTime = SetOperationTimeValue;
                    Properties.Settings.Default.Save();
                }
                //Device.GetOperationTime();
            }));
        }
        void ReceivedSetPowerGainResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ReceivedPowerGainIndex = -1;
                ShowMessage("Set power succeed.");
                if (!string.IsNullOrEmpty(SetPowerGainValue))
                {
                    Properties.Settings.Default.PowerGain = SetPowerGainValue;
                    Properties.Settings.Default.Save();
                    //string val = SetPowerGainValue + " dB";
                    //cmbPowerGain.SelectedValue = val;
                    //cmbReadPowerGain.SelectedValue = val;
                    //cmbWritePowerGain.SelectedValue = val;
                    //cmbLockPowerGain.SelectedValue = val;
                }
                Device.GetPower();
            }));
        }
        void ReceivedSetInventoryModeResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set inventory mode succeed.");
                if (!string.IsNullOrEmpty(SetContinuousModeValue))
                {
                    Properties.Settings.Default.ContinuousMode = SetContinuousModeValue;
                    Properties.Settings.Default.Save();
                }
            }));
            Device.GetInventoryMode();
        }
        void ReceivedSetSessionFlagResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set session flag succeed.");
                if (!string.IsNullOrEmpty(SetSessionFlagValue))
                {
                    Properties.Settings.Default.SessionFlag = SetSessionFlagValue;
                    Properties.Settings.Default.Save();
                }
            }));
        }
        void ReceivedSetRssiOnOffResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set rssi on off succeed.");
                if (!string.IsNullOrEmpty(SetRssiValue))
                {
                    Properties.Settings.Default.ReportRssi = SetRssiValue;
                    Properties.Settings.Default.Save();
                }
                Device.GetRssiOnOff();
            }));
        }
        void ReceivedSetSessionResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set session succeed.");
                if (!string.IsNullOrEmpty(SetSessionValue))
                {
                    Properties.Settings.Default.Session = SetSessionValue;
                    Properties.Settings.Default.Save();
                }
            }));
        }
        void ReceivedSessionFlag(int sessionFlag)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                ShowMessage("Received session flag.");
                cmbSessionFlag.SelectedIndex = sessionFlag;
            }));
            //Log.Info($"Received session flag: {sessionFlag}");
        }
        void ReceivedRssiOnOff(Types.RssiOnOff onOff)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Received report rssi.");
                dgTags.Columns[5].Visibility = dgTags.Columns[6].Visibility = dgTags.Columns[7].Visibility = onOff == Types.RssiOnOff.On ? Visibility.Visible : Visibility.Hidden;
                cbReportRssi.IsChecked = cbReadReportRssi.IsChecked = cbWriteReportRssi.IsChecked = cbLockReportRssi.IsChecked = onOff == Types.RssiOnOff.On;
            }));
            //Log.Info($"Received report rssi: {onOff}");
        }
        void ReceivedInventoryMode(Types.InventoryMode mode)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                ShowMessage("Received inventory mode.");
                cbContinuousMode.IsChecked = mode == Types.InventoryMode.Continuous;
            }));
            //Log.Info($"Received inventory mode: {mode}");
        }
        void ReceivedOperationTime(int operationTime)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                ShowMessage("Received operation time.");
                txtOperationTime.Text = txtReadOperationTime.Text = txtWriteOperationTime.Text = txtLockOperationTime.Text = operationTime.ToString();
            }));
            //Log.Info($"Received operation time: {operationTime}");
        }
        void ReceivedKillResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Kill tag succeed.");
                //SwitchButtonContent();
            }));
        }
        void SwitchButtonContent()
        {
            switch (actionType)
            {
                case ActionType.Read:
                    btnRead.Content = "Read";
                    break;
                case ActionType.Write:
                    btnWrite.Content = "Write";
                    break;
                case ActionType.Lock:
                    btnLock.Content = "Lock";
                    break;
                case ActionType.Unlock:
                    btnUnLock.Content = "Unlock";
                    break;
                case ActionType.Permalock:
                    btnPermalock.Content = "Permalock";
                    break;
                case ActionType.Kill:
                    btnKill.Content = "Kill";
                    break;
                case ActionType.Set_Access_Password:
                    btnSetAccessPwd.Content = "Set Access Password";
                    break;
                case ActionType.Set_Kill_Password:
                    btnSetKillPwd.Content = "Set Kill Password";
                    break;
            }
        }
        void ReceivedLockResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage(Enum.GetName(typeof(ActionType), actionType) + " succeed.");
                //SwitchButtonContent();
            }));
        }
        void ReceivedWriteResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Write tag succeed.");
                SwitchButtonContent();
            }));
        }
        void ReceivedReadResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Read tag succeed.");
            }));
        }
        void ReceivedSetPasswordResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set password succeed.");
            }));
        }
        void ReceivedSetEPCMaskResult()
        {
            Properties.Settings.Default.MaskOffset = SetMaskOffsetValue.ToString();
            Properties.Settings.Default.MaskData = SetMaskDataValue;
            Properties.Settings.Default.MaskLength = SetMaskLengthValue.ToString();
            Properties.Settings.Default.Save();
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Mask saved succeed.");
                if (_LEDStatus != LEDStatus.On) Device.GetEPCMaskData();
            }));
        }
        void ReceivedClearEPCMaskResult()
        {
            Properties.Settings.Default.MaskOffset = "";
            Properties.Settings.Default.MaskData = "";
            Properties.Settings.Default.MaskLength = "";
            Properties.Settings.Default.Save();
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Clear EPC mask data succeed.");
                cmbMaskOffset.SelectedIndex = -1;
                txtMaskData.Text = "";
                cmbMaskLength.SelectedIndex = -1;
            }));
        }
        void ReceivedEPCMaskData(int maskOffset, string maskData, int maskLength)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                cmbMaskOffset.SelectedValue = maskOffset;
                txtMaskData.Text = maskData;
                cmbMaskLength.SelectedValue = maskLength;
            }));
        }
        Encoding encoding = Encoding.ASCII;
        void ReceivedBarcodeData(byte[] barcodeData, string typeName)
        {
            try
            {
                // Device.Stop();
                Task.Run(() =>
                {
                    Dispatcher.Invoke((Action)(() =>
                    {
                        string value;
                        char[] chr = new char[] { '\r', '\n' };
                        value = encoding.GetString(barcodeData).TrimEnd(chr);
                        var query = (from t in BarcodeList
                                     where t.BarcodeData == value && t.TypeName == typeName
                                     select BarcodeList.IndexOf(t)).ToList();
                        if (query.Count > 0)
                        {
                            foreach (int index in query)
                            {
                                BarcodeList[index].Count += 1;
                            }
                        }
                        else
                        {
                            Barcode barcode = new Barcode()
                            {
                                Index = BarcodeList.Count + 1,
                                BarcodeData = value,
                                TypeName = typeName,
                                Count = 1,
                            };
                            BarcodeList.Add(barcode);
                        }
                        dgBarcodes.ItemsSource = null;
                        dgBarcodes.ItemsSource = BarcodeList;
                        //dgBarcodes.DataContext = BarcodeList;
                        //dgBarcodes.Items.Refresh();
                        //dgBarcodes.Items.Add(new DataGridRow() { Item = dGBarcode });
                        lblBacodeCnt.Content = BarcodeList.Count.ToString();
                    }));
                });
            }
            catch (Exception ex)
            {
                Console.WriteLine($"ReceivedBarcode:{ex.Message}");
            }
        }
        double preOffet = 0;
        private void dataGrid_ScrollChanged(object sender, ScrollChangedEventArgs e)
        {
            try
            {
                DataGrid dg = (DataGrid)sender;
                double suffOffet = e.VerticalOffset;
                if (e.VerticalOffset >= 0 && suffOffet >= preOffet)
                {
                    dg.Items.MoveCurrentToLast();
                    //if (suffOffet - preOffet >= 1) dg.Items.MoveCurrentToNext();
                    //else dg.Items.MoveCurrentToLast();
                    if (dg.Items.CurrentItem != null)
                    {
                        var item = dg.Items.CurrentItem;
                        dg.ScrollIntoView(item);
                        dg.UpdateLayout();
                    }
                }
                preOffet = suffOffet;
            }
            catch (Exception ex)
            {
                Console.WriteLine($"dataGrid_ScrollChanged:{ex.Message}");
            }
        }
        //private void dgTags_ScrollChanged(object sender, ScrollChangedEventArgs e)
        //{
        //    try
        //    {
        //        DataGrid dg = (DataGrid)sender;
        //        if (dg.Items.CurrentItem != null)
        //        {
        //            dg.Items.MoveCurrentToLast();
        //            var item = dg.Items.CurrentItem;
        //            dg.ScrollIntoView(item);
        //            dg.UpdateLayout();
        //        }
        //    }
        //    catch (Exception ex)
        //    {
        //        Console.WriteLine($"dgTags_ScrollChanged:{ex.Message}");
        //    }
        //}
        void ReceivedStartBarcodeScanResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                if (cmbRescanTime.SelectedIndex > 0) return;
                ShowMessage("Start Barcode Scan succeed.");
            }));
        }
        void ReceivedStopResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                Console.WriteLine("Stop.......");
                SwitchButtonContent();
                if (btnScanBarcode.Content.ToString() == "Stop Scan" && cmbRescanTime.SelectedIndex > 0) return;
                ShowMessage("Stop succeed.");
                btnScanBarcode.Content = "Start Scan";
                btnInventory.Content = "Start";
                btnRead.Content = "Read";
                EnabledInventoryItems(true);
                if (TRescanBarcodeTimer != null)
                {
                    TRescanBarcodeTimer.Dispose();
                    TRescanBarcodeTimer = null;
                }
                if (STimerInventory != null) STimerInventory.Stop();
                if (InventoryTimer != null)
                {
                    ObservableCollection<Tag> tmp = new ObservableCollection<Tag>(TagList);
                    InventoryTimer?.Dispose();
                    InventoryTimer = null;
                    dgTags.ItemsSource = null;
                    dgTags.ItemsSource = tmp;
                    lblTagCount.Content = tmp.Count;
                    lblReadCount.Content = tmp.Count > 0 ? tmp[0].TotalCount : 0;//readTagCount;
                }
            }));
        }
        void ReceivedSetBarcodeTimeoutResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set Barcode Timeout succeed.");
                Properties.Settings.Default.BarcodeTimeout = SetBarcodeTimeoutValue;
                Properties.Settings.Default.Save();
            }));
        }
        void ReceivedSetAutoOffTimeResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set Auto Off Time succeed.");
                Properties.Settings.Default.AutoOffTime = SetAutoOffTimeValue;
                Properties.Settings.Default.Save();
            }));
        }
        void ReceivedSetAutoOffSleepResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set Sleep Time succeed.");
                Properties.Settings.Default.SleepTime = SetSleepTimeValue;
                Properties.Settings.Default.Save();
            }));
        }
        void ReceivedSetIdleTimeResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set Idle Time succeed.");
                Properties.Settings.Default.IdleTime = SetIdleTimeValue;
                Properties.Settings.Default.Save();
            }));
        }
        void ReceivedSetBuzzerResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set Buzzer succeed.");
                Properties.Settings.Default.Buzzer = SetBuzzerValue;
                Properties.Settings.Default.Save();
            }));
        }
        void ReceivedSetFrequencyResult()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Set Frequency succeed.");
                if (SetFrequencyAutomaticValue != null)
                {
                    Properties.Settings.Default.FrequencyAutomatic = SetFrequencyAutomaticValue;
                    if (SetFrequencyAutomaticValue == "0")
                    {
                        Properties.Settings.Default.LBTChannel = SetFrequencyAssignedValue;
                    }
                    else
                    {
                        ClearLBTChannels();
                    }
                    Properties.Settings.Default.Save();
                }
            }));
        }
        //bool? RestoreSuccessed = null;
        void ReceivedRestoreFactorySettings()
        {
            GetOption();
            Dispatcher.Invoke((Action)(() =>
            {
                ShowMessage("Restore Factory Settings succeed.");
                Properties.Settings.Default.DefaultLinkProfile = _defaultLinkProfile.ToString();
                Properties.Settings.Default.BatteryInterval = _batteryInterval.ToString();
            }));
        }

        int[] _receivedFrequencyArr = { };
        void ReceivedFrequency(int[] frequencyArr)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                if (frequencyArr.Length == 0)
                {
                    cbFrequencyAuto.IsChecked = true;
                    lblLBTChannel.Visibility = Visibility.Hidden;
                    lvBTChannel.Visibility = Visibility.Hidden;
                }
                else
                {
                    _receivedFrequencyArr = frequencyArr;
                    if (freqArr != null && freqArr.Length > 0)
                    {
                        lblLBTChannel.Visibility = Visibility.Visible;
                        //cmbLBTChannel.Visibility = Visibility.Visible;
                        lvBTChannel.Visibility = Visibility.Visible;
                        KVFrequencyList = new List<Bind>();
                        bool isChecked = false;
                        for (int i = 0; i < freqArr.Length; i++)
                        {
                            if (Array.IndexOf(_receivedFrequencyArr, i) >= 0) isChecked = true;
                            else isChecked = false;
                            KVFrequencyList.Add(new Bind() { Id = i, Value = freqArr[i], isChecked = isChecked });
                        }
                        //cmbLBTChannel.ItemsSource = KVFrequencyList;
                        lvBTChannel.ItemsSource = KVFrequencyList;
                    }
                }
                Properties.Settings.Default.FrequencyAutomatic = cbFrequencyAuto.IsChecked == true ? "1" : "0";
            }));
            //Log.Info($"Received frequency: {data}");
        }
        void ReceivedFrequencyList(string[] arr)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                freqArr = arr;
                KVFrequencyList = new List<Bind>();
                bool isChecked = false;
                for (int i = 0; i < arr.Length; i++)
                {
                    if (_receivedFrequencyArr.Length > 0 && Array.IndexOf(_receivedFrequencyArr, i) >= 0) isChecked = true;
                    else isChecked = false;
                    KVFrequencyList.Add(new Bind() { Id = i, Value = arr[i], isChecked = isChecked });
                }
                //cmbLBTChannel.ItemsSource = KVFrequencyList;
                lvBTChannel.ItemsSource = KVFrequencyList;
            }));
        }
        void ReceivedBarcodeTimeout(int barcodeTimeout)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                if (barcodeTimeout > 0)
                {
                    txtBarcodeTimeout.Text = barcodeTimeout.ToString(); 
                    string saved = Properties.Settings.Default.BarcodeTimeout;
                    if (!string.IsNullOrEmpty(saved) && int.TryParse(saved, out _) && barcodeTimeout.ToString() != saved) SetBarcodeTimeout(saved);
                }
                else
                {
                    SetBarcodeTimeout("4");
                    Thread.Sleep(100);
                    Device.GetBarcodeTimeout();
                }
                //if (barcodeTimeout != 0) Properties.Settings.Default.BarcodeTimeout = barcodeTimeout.ToString();
            }));
            //Log.Info($"Received barcode scan time: {data}");
        }
        void ReceivedBaudRateList(string[] list)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                cmbOptionBaudRate.ItemsSource = list;
            }));
        }
        void ReceivedBaudRate(int data)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                cmbOptionBaudRate.SelectedIndex = data;
                Properties.Settings.Default.BaudRate = data.ToString();
            }));
            //Log.Info($"Received baud rate: {data}");
        }
        void ReceivedLinkProfile(int data)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                cmbLinkProfile.Text = data.ToString();
                Properties.Settings.Default.LinkProfile = data.ToString();
            }));
            //Log.Info($"Received link profile: {data}");
        }
        void ReceivedBattery(int battery)
        {
            Task.Run(() =>
            {
                if (_connectTypes == ConnectTypes.Bluetooth)
                {
                    rcvBattery = battery;
                    resetEvent.Reset();
                }
            });
            Dispatcher.BeginInvoke((Action)(() =>
            {
                ShowMessage($"battery:{battery}");
                if (battery == 1)
                {
                    imgBattery.Source = new BitmapImage(new Uri("Resources/battery1.png", UriKind.RelativeOrAbsolute));
                }
                else if (battery == 2)
                {
                    imgBattery.Source = new BitmapImage(new Uri("Resources/battery2.png", UriKind.RelativeOrAbsolute));
                }
                else if (battery == 3)
                {
                    imgBattery.Source = new BitmapImage(new Uri("Resources/battery3.png", UriKind.RelativeOrAbsolute));
                }
                else if (battery == 4)
                {
                    imgBattery.Source = new BitmapImage(new Uri("Resources/battery4.png", UriKind.RelativeOrAbsolute));
                }
                else
                {
                    imgBattery.Source = new BitmapImage(new Uri("Resources/battery0.png", UriKind.RelativeOrAbsolute));
                }
            }));
        }
        void ReceivedBuzzer(Types.BuzzerStatus data)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                cmbBuzzer.SelectedIndex = (int)data;
            }));
            //Log.Info($"Received buzzer: {data}");
        }
        void ReceivedAutoOffSleep(int data)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                txtSleepTime.Text = data.ToString();
                Properties.Settings.Default.SleepTime = data.ToString();
            }));
            //Log.Info($"Received auto off sleep: {data}");
        }
        void ReceivedAutoOffTime(int data)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                txtAutoOffTime.Text = data.ToString();
                Properties.Settings.Default.AutoOffTime = data.ToString();
            }));
            //Log.Info($"Received auto off time: {data}");
        }
        void ReceivedQValue(int data)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                cmbQValue.SelectedIndex = data;
                Properties.Settings.Default.QValue = data.ToString();
            }));
            //Log.Info($"Received QValue: {data}");
        }
        void ReceivedQMaxValue(int data)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                cmbMaxQ.SelectedIndex = data;
            }));
            //Log.Info($"Received QMaxValue: {data}");
        }
        void ReceivedQMinValue(int data)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                cmbMinQ.SelectedIndex = data;
            }));
            //Log.Info($"Received QMinValue: {data}");
        }
        void ReceivedIdleTime(int data)
        {
            Dispatcher.BeginInvoke((Action)(() =>
            {
                txtIdle.Text = data.ToString();
                Properties.Settings.Default.IdleTime = data.ToString();
            }));
            //Log.Info($"Received idle time: {data}");
        }
        void ReceivedHardVersion(string data)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                lblHardVer.Content = data;
            }));
        }
        void ReceivedFirmwareVersion(string data)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                lblFirmwareVer.Content = data;
            }));
        }
        void ReceivedRegionCountry(string data)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                lblRegionCoutry.Content = data;
                lblFrequencyAutoTitle.Visibility =
                cbFrequencyAuto.Visibility =
                lblLBTChannel.Visibility =
                lvBTChannel.Visibility = data.ToLower().IndexOf("japan") >= 0 ? Visibility.Visible : Visibility.Hidden;
            }));
        }
        void ReceivedSerialNumber(string data)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                lblSN.Content = data;
            }));
        }
        void ReceivedRFIDModuleVersion(string data)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                lblUhfHf.Content = data;
            }));
        }
        #endregion delegate functions

        void GetInformations()
        {
            Device.GetFirmwareVersion();
            Thread.Sleep(100);
            Device.GetHardVersion();
            Thread.Sleep(100);
            Device.GetRegionCountry();
            Thread.Sleep(100);
            Device.GetSerialNumber();
            Thread.Sleep(100);
            Device.GetRFIDModuleVersion();
        }
        void GetOption()
        {
            Thread.Sleep(100);
            Device.GetBuzzer();
            Thread.Sleep(200);
            Device.GetIdleTime();
            Thread.Sleep(100);
            Device.GetAutoOffSleep();
            Thread.Sleep(100);
            Dispatcher.BeginInvoke((Action)(() =>
            {
                txtBatteryInterval.Text = Properties.Settings.Default.BatteryInterval;
            }));
            Device.GetAutoOffTime();
            Thread.Sleep(100);
            Device.GetQValue();
            Thread.Sleep(100);
            Device.GetQMaxValue();
            Thread.Sleep(100);
            Device.GetQMinValue();
            Thread.Sleep(100);
            Device.GetLinkProfile();
            Thread.Sleep(100);
            Dispatcher.BeginInvoke((Action)(() =>
            {
                int selIndex = -1;
                int.TryParse(Properties.Settings.Default.DefaultLinkProfile, out selIndex);
                cmbDefaultLinkProfile.SelectedIndex = selIndex;
            }));
            Device.GetBaudRate();
            Thread.Sleep(100);
            Device.GetFrequencyList();
            Thread.Sleep(100);
            Device.GetFrequency();
            Thread.Sleep(100);
            Device.GetBarcodeTimeout();
        }
        private void tabInventory_Clicked(object sender, RoutedEventArgs e)
        {
            if (connectStatus == ConnectStatus.Connected)
            {
                Task.Run(() =>
                {
                    GetRFIDSettings();
                });
                if (cbDisplayPC.IsChecked == true) dgTags.Columns[4].Visibility = Visibility.Visible;
                else dgTags.Columns[4].Visibility = Visibility.Hidden;
            }
        }
        private void tabItemInfo_Clicked(object sender, RoutedEventArgs e)
        {
            if (connectStatus == ConnectStatus.Connected)
            {
                ClearInformation();
                Task.Run(() =>
                {
                    GetInformations();
                });
            }
        }
        private void tabItemMask_Clicked(object sender, RoutedEventArgs e)
        {
            if (connectStatus == ConnectStatus.Connected)
            {
                Device.GetEPCMaskData();
            }
        }
        private void cbFrequencyAuto_Click(object sender, RoutedEventArgs e)
        {
            CheckBox ck = (CheckBox)sender;
            lblLBTChannel.Visibility = ck.IsChecked == true ? Visibility.Hidden : Visibility.Visible;
            lvBTChannel.Visibility = ck.IsChecked == true ? Visibility.Hidden : Visibility.Visible;
        }
        private void tabItemOption_Clicked(object sender, RoutedEventArgs e)
        {
            if (connectStatus == ConnectStatus.Connected)
            {
                ClearOption();
                Task.Run(() =>
                {
                    GetOption();
                });
            }
        }
        private void btnMaskSave_Click(object sender, RoutedEventArgs e)
        {
            if (txtMaskData.Text == "")
            {
                ShowMessage("Mask data can not be null.");
            }
            else if (txtMaskData.Text.Length / 4 != Convert.ToInt32(cmbMaskLength.SelectedValue) / 16)
            {
                ShowMessage("Mask length is not valid.");
            }
            else
            {
                Device.ClearEPCMaskData();
                Thread.Sleep(50);
                SetMaskOffsetValue = Convert.ToInt32(cmbMaskOffset.SelectedValue);
                SetMaskDataValue = txtMaskData.Text.Trim();
                SetMaskLengthValue = Convert.ToInt32(cmbMaskLength.SelectedValue);
                Device.SetEPCMaskData(SetMaskOffsetValue, SetMaskDataValue, SetMaskLengthValue);
            }
        }
        private void btnMaskClear_Click(object sender, RoutedEventArgs e)
        {
            Device.ClearEPCMaskData();
        }
        private void btnConnect_Click(object sender, RoutedEventArgs e)
        {
            if (ConnectBTDeviceTimer.Enabled)
            {
                ConnectBTDeviceTimer.Stop();
                btnSearch.IsEnabled = btnBTConnect.IsEnabled = true;
            }
            if (cmbPorts.SelectedIndex < 0 || cmbBaundrate.SelectedIndex < 0)
            {
                ShowMessage("Connect parameters can not be null.");
                return;
            }
            UInt32 ret = Device.Connect(cmbPorts.Text, Convert.ToInt32(cmbBaundrate.Text));
            if (ret == 0)
            {
                _connectTypes = ConnectTypes.COM;
                ShowMessage("Connect succeed.");
                WhenConnected();
            }
            else
            {
                ShowMessage("Connect Failed.");
            }
        }
        void WhenConnected(bool connResult = true)
        {
            if (connResult)
            {
                connectStatus = ConnectStatus.Connected;
                ShowMessage("Connect succeed.");
                if (_connectTypes == ConnectTypes.COM)
                {
                    Properties.Settings.Default.BluetoothDevice = null;
                    ConnectBTDeviceTimer.Stop();
                }
                else Properties.Settings.Default.BluetoothDevice = currentBluetoothDevice;
                if (SettingBatteryInterval > 0)
                {
                    SetBatteryInterverTimer(SettingBatteryInterval);
                    GetBatteryTimer.Start();
                }
                ClearInformation();
                ClearOption();
                EnableControllers();
                //GetTemperatureTimer.Start();
                Task.Run(() =>
                {
                    Device.GetPowerRange();
                    Thread.Sleep(100);
                    Device.GetBaudRateList();
                    SetRFIDSetting();
                    Thread.Sleep(100);
                    GetRFIDSettings();
                    Thread.Sleep(100);
                    Device.GetBattery();
                    Thread.Sleep(100);
                    GetInformations();
                    Thread.Sleep(100);
                    //InitOption();
                    //Thread.Sleep(100);
                    GetOption();
                    //Thread.Sleep(100);
                    //InitMask();
                });
            }
            else {
                btnSearch.IsEnabled = btnBTConnect.IsEnabled = true;
                ShowMessage("Connect failed.");
            }
        }
        void InitMask()
        {
            Device.ClearEPCMaskData();
            string sOffset = Properties.Settings.Default.MaskOffset;
            string data = Properties.Settings.Default.MaskData;
            string sLength = Properties.Settings.Default.MaskLength;
            if (!string.IsNullOrEmpty(sOffset) &&
                !string.IsNullOrEmpty(data) &&
                !string.IsNullOrEmpty(sLength))
            {
                int offset, length;
                if (int.TryParse(sOffset, out offset) && int.TryParse(sLength, out length))
                {
                    if (data.Length / 4 == length / 16)
                    {
                        Thread.Sleep(100);
                        Device.SetEPCMaskData(offset, data, length);
                    }
                }
            }
        }
        
        void GetRFIDSettings()
        {
            Thread.Sleep(200);
            Device.GetPower();
            Thread.Sleep(200);
            Device.GetSession();
            Thread.Sleep(200);
            Device.GetSessionFlag();
            Thread.Sleep(200);
            Device.GetOperationTime();
            Thread.Sleep(200);
            Device.GetInventoryMode();
            Thread.Sleep(200);
            Device.GetRssiOnOff();
            Thread.Sleep(200);
        }
        void SetRFIDSetting()
        {
            //if (Properties.Settings.Default.PowerGain != "") SetPowerGain(Properties.Settings.Default.PowerGain);
            //Thread.Sleep(100);
            //if (Properties.Settings.Default.Session != "") SetSession(Properties.Settings.Default.Session);
            //Thread.Sleep(100);
            //if (Properties.Settings.Default.SessionFlag != "") SetSessionFlag(Properties.Settings.Default.SessionFlag);
            //Thread.Sleep(100);
            //if (Properties.Settings.Default.OperationTime != "") SetOperationTime(Properties.Settings.Default.OperationTime);
            //Thread.Sleep(100);
            Device.SetInventoryMode(Types.InventoryMode.Continuous);
            //if (Properties.Settings.Default.ContinuousMode != "") SetContinuousMode(Properties.Settings.Default.ContinuousMode);
            //Thread.Sleep(100);
            //if (Properties.Settings.Default.ReportRssi != "") SetReportRssi(Properties.Settings.Default.ReportRssi);
        }
        private void btnDisconnect_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (tbRight.SelectedIndex == 0)
                {
                    if (btnInventory.Content.ToString() == "Stop") btnStartInventory_Click(null, null);
                    if (btnLED.Content.ToString() == "LED OFF") btnLED_Click(null, null);
                }
                if (tbRight.SelectedIndex == 1 && btnScanBarcode.Content.ToString() == "Stop Scan") btnScanBarcode_Click(null, null);
                UInt32 ret = Device.Disconnect();
                if (ret == 0)
                {
                    WhenDisconnected();
                }
                else
                {
                    ShowMessage("Disconnect failed.");
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("btnDisconnect_Click: " + ex.Message);
            }
        }
        void WhenDisconnected()
        {
            ShowMessage("Disconnect succeed.");
            lblWarning.Visibility = Visibility.Hidden;
            connectStatus = ConnectStatus.Disconnected;
            _connectTypes = ConnectTypes.None;
            imgBattery.Source = new BitmapImage(new Uri("Resources/battery0.png", UriKind.RelativeOrAbsolute));
            EnableControllers();
            if (GetBatteryTimer != null) GetBatteryTimer.Stop();
            //if (GetTemperatureTimer.Enabled) GetTemperatureTimer.Stop();
            if (!_isBTDisConnClick) ConnectBTDeviceTimer.Start();
            else
            {
                ConnectBTDeviceTimer.Stop();
                Properties.Settings.Default.BluetoothDevice = null;
                _isBTDisConnClick = false;
            }
            lblWorkTemperature.Content = "- ℃";
        }
        private void btnScanBarcode_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (btnScanBarcode.Content.ToString() == "Start Scan")
                {
                    encoding = rbASCII.IsChecked == true ? Encoding.ASCII : rbUTF8.IsChecked == true ? Encoding.UTF8 : Encoding.GetEncoding("Shift-JIS");
                    btnScanBarcode.Content = "Stop Scan";
                    if (cmbRescanTime.SelectedIndex == 0) Device.StartBarcodeScan();
                    else RescanBarcode();
                }
                else
                {
                    //if(RescanBarcodeTimer.Enabled) RescanBarcodeTimer.Stop();
                    if (TRescanBarcodeTimer != null)
                    {
                        TRescanBarcodeTimer.Dispose();
                        TRescanBarcodeTimer = null;
                    }
                    Device.Stop();
                    btnScanBarcode.Content = "Start Scan";
                }
            }
            catch (Exception ex)
            {
                ShowMessage($"Scan Barcode: {ex.Message}");
            }

        }
        void RescanTimeDo(object obj)
        {
            Device.StartBarcodeScan();
        }
        void RescanBarcode()
        {
            try
            {
                TRescanBarcodeTimer = new System.Threading.Timer(new System.Threading.TimerCallback(RescanTimeDo), null, 0, int.Parse(cmbRescanTime.Text.Replace(" ms", "")));
                //RescanBarcodeTimer.Interval = int.Parse(cmbRescanTime.Text.Replace(" ms", ""));
                //RescanBarcodeTimer.Enabled = true;
                //RescanBarcodeTimer.Start();
            }
            catch (Exception ex)
            {
                ShowMessage($"RescanBarcode:{ex.Message}");
            }
        }
        private void btnRestoreFactorySettings_Click(object sender, RoutedEventArgs e)
        {
            ClearOption();
            Properties.Settings.Default.DefaultLinkProfile = _defaultLinkProfile.ToString();
            Properties.Settings.Default.BatteryInterval = _batteryInterval.ToString();
            Device.RestoreFactorySettings();
        }
        bool IsNumeric(string data, int? min, int? max, string name)
        {
            Regex rgx = new Regex("^[0-9]+$");
            if (rgx.IsMatch(data) == false)
            {
                ShowMessage($"{name}: Please enter a numeric value.");
                return false;
            }
            else
            {
                int val = int.Parse(data);
                if (max.HasValue && min.HasValue && (val > max || val < min))
                {
                    ShowMessage($"{name}: Value out of range({min}~{max}).");
                    return false;
                }
            }
            return true;
        }
        void SetBuzzer()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Buzzer";
                if (cmbBuzzer.SelectedIndex >= 0)
                {
                    SetBuzzerValue = cmbBuzzer.SelectedIndex.ToString();
                    Device.SetBuzzer((Types.BuzzerStatus)cmbBuzzer.SelectedIndex);
                    //Log.Info($"Set Buzzer Value: {SetBuzzerValue}");
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter can not be null.");
                }
            }));
        }
        void SetIdleTime(string txtIdle)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Idle Time";
                if (txtIdle != "" && IsNumeric(txtIdle, 0, 65535, name))
                {
                    Device.SetIdleTime(txtIdle);
                    SetIdleTimeValue = txtIdle;
                    //Log.Info($"Set Idle Time: {txtIdle}");
                }
                else
                {
                    if (txtIdle == "") ShowMessage($"Set {name}: Parameter can not be null.");
                }
            }));
        }
        void SetSleepTime(string txtSleepTime)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Sleep Time";
                if (txtSleepTime != "" && IsNumeric(txtSleepTime, 0, 1800, name))
                {
                    Device.SetAutoOffSleep(txtSleepTime);
                    SetSleepTimeValue = txtSleepTime;
                    //Log.Info($"Set Sleep Time Value: {txtSleepTime}");
                }
                else
                {
                    if (txtSleepTime == "") ShowMessage($"Set {name}: Parameter can not be null.");
                }
            }));
        }
        void SetAutoOffTime(string txtAutoOffTime)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Auto Off Time";
                if (txtAutoOffTime != "" && IsNumeric(txtAutoOffTime, 0, 1800, name))
                {
                    SetAutoOffTimeValue = txtAutoOffTime;
                    Device.SetAutoOffTime(txtAutoOffTime);
                    //Log.Info($"Set Auto Off Time: {txtAutoOffTime}");
                }
                else
                {
                    if (txtAutoOffTime == "") ShowMessage($"Set {name}: Parameter can not be null.");
                }
            }));
        }
        void SetQValue(string QValue)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "QValue";
                int index = -1;

                if (string.IsNullOrEmpty(QValue))
                {
                    ShowMessage($"Set {name}: Parameter can not be null.");
                }
                else if (int.TryParse(QValue, out index) && index >= 0)
                {
                    Device.SetQValue(QValue);
                    SetQValueValue = QValue;
                    //Log.Info($"Set QValue: {QValue}");
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter error.(value={QValue})");
                }
            }));
        }
        void SetLinkProfile(string linkProfile)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Link Profile";
                int index = -1;
                if (string.IsNullOrEmpty(linkProfile))
                {
                    ShowMessage($"Set {name}: Parameter can not be null.");
                }
                else if (int.TryParse(linkProfile, out index) && index >= 0)
                {
                    Device.SetLinkProfile(linkProfile);
                    SetLinkProfileValue = linkProfile;
                    //Log.Info($"Set Link Profile: {linkProfile}");
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter error.(value={linkProfile})");
                }
            }));
        }
        void SetBarcodeTimeout(string txtBarcodeTimeout)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Barcode Timeout";
                if (txtBarcodeTimeout != "")
                {
                    if (IsNumeric(txtBarcodeTimeout, 4, 300, name))
                    {
                        Device.SetBarcodeTimeout(txtBarcodeTimeout);
                        SetBarcodeTimeoutValue = txtBarcodeTimeout;
                        //Log.Info($"Set Barcode Timeout: {txtBarcodeTimeout}");
                    }
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter can not be null.");
                }
            }));
        }
        void SetFrequencyAssigned()
        {
            Dispatcher.Invoke((Action)(() =>
            {
                int[] arr = GetLBTChannels();
                if (arr.Length == 0) ShowMessage("Please select the LBT Channel");
                else Device.SetFrequencyAssigned(arr);
            }));
        }
        int[] GetLBTChannels()
        {
            List<int> list = new List<int>();
            foreach (Bind item in lvBTChannel.Items)
            {
                if (item.isChecked)
                {
                    list.Add((byte)(item.Id));
                }
            }
            return list.ToArray();
        }
        void ClearLBTChannels()
        {
            KVFrequencyList = new List<Bind>();
            for (int i = 0; i < freqArr.Length; i++)
            {
                KVFrequencyList.Add(new Bind() { Id = i, Value = freqArr[i], isChecked = false });
            }
            lvBTChannel.ItemsSource = null;
            lvBTChannel.ItemsSource = KVFrequencyList;
        }
        void SetBaudRate(string baudRateIndex)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Baud Rate";
                int index = -1;
                if (string.IsNullOrEmpty(baudRateIndex))
                {
                    ShowMessage($"Set {name}: Parameter can not be null.");
                }
                else if (int.TryParse(baudRateIndex, out index) && index >= 0)
                {
                    Device.SetBaudRate(index);
                    //Log.Info($"Set Baud Rate: {index}");
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter error.(value={baudRateIndex})");
                }
            }));
        }
        string SetPowerGainValue;
        void SetPowerGain(string powerGainArgs)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Power Gain";
                double powerGain = 0;
                if (double.TryParse(powerGainArgs, out powerGain))
                {
                    if (powerGain >= 0)
                    {
                        Device.SetPowerGain((int)(powerGain * 10));
                        SetPowerGainValue = powerGainArgs;
                    }
                    else
                    {
                        ShowMessage($"Set {name}: Parameter value error.(value={powerGainArgs})");
                    }
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter convert to double error.(value={powerGainArgs})");
                }
            }));
        }
        void SetSession(string session)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Session";
                int sessionArgs = -1;
                if (int.TryParse(session, out sessionArgs))
                {
                    if (sessionArgs >= 0)
                    {
                        Device.SetSession(sessionArgs);
                        SetSessionValue = sessionArgs.ToString();
                    }
                    else
                    {
                        ShowMessage($"Set {name}: Parameter value error.(value={session})");
                    }
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter convert to int error.(value={session})");
                }
            }));
        }
        void SetSessionFlag(string sessionFlag)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Session Flag";
                int sessionFlagArgs = -1;
                if (int.TryParse(sessionFlag, out sessionFlagArgs))
                {
                    if (sessionFlagArgs >= 0)
                    {
                        Device.SetSessionFlag(sessionFlagArgs);
                        SetSessionFlagValue = sessionFlagArgs.ToString();
                    }
                    else
                    {
                        ShowMessage($"Set {name}: Parameter value error.(value={sessionFlag})");
                    }
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter convert to int error.(value={sessionFlag})");
                }
            }));
        }
        string SetOperationTimeValue;
        void SetOperationTime(string value)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Operation Time";
                if (!string.IsNullOrEmpty(value) && IsNumeric(value, 0, 1800, name))
                {
                    Device.SetOperationTime(value);
                    SetOperationTimeValue = value;
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter can not be null.");
                }
            }));
        }
        void SetContinuousMode(string inventoryMode)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                string name = "Continuous Mode";
                if (inventoryMode == "0" || inventoryMode == "1")
                {
                    Types.InventoryMode mode = (Types.InventoryMode)int.Parse(inventoryMode);
                    Device.SetInventoryMode(mode);
                    SetContinuousModeValue = inventoryMode;
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter error.(value={inventoryMode})");
                }
            }));
        }
        string SetReportRssiValue;
        void SetReportRssi(string reportRssi)
        {
            Dispatcher.Invoke((Action)(() => {
                string name = "Report Rssi";
                if (reportRssi == "0" || reportRssi == "1")
                {
                    Types.RssiOnOff rssiOnOff = (Types.RssiOnOff)int.Parse(reportRssi);
                    Device.SetRssiOnOff(rssiOnOff);
                    SetReportRssiValue = reportRssi;
                }
                else
                {
                    ShowMessage($"Set {name}: Parameter error.(value={reportRssi})");
                }
            }));
        }
        private void btnBarcodeScanClear_Click(object sender, RoutedEventArgs e)
        {
            lblBacodeCnt.Content = "";
            dgBarcodes.ItemsSource = null;
            dgBarcodes.Items.Clear();
            txtLog.Text = "";
            BarcodeList.Clear();
        }
        string SetFrequencyAutomaticValue, SetFrequencyAssignedValue,
            SetBuzzerValue, SetIdleTimeValue, SetSleepTimeValue,
            SetAutoOffTimeValue, SetQValueValue, SetLinkProfileValue,
            SetBarcodeTimeoutValue, SetBaudRateValue, SetMaskDataValue;
        int SetMaskOffsetValue, SetMaskLengthValue;
        private void btnOptionSave_Click(object sender, RoutedEventArgs e)
        {
            btnOptionSave.IsEnabled = false;
            SetIdleTimeValue = txtIdle.Text;
            SetSleepTimeValue = txtSleepTime.Text;
            SetAutoOffTimeValue = txtAutoOffTime.Text;
            SetQValueValue = cmbQValue.SelectedIndex.ToString();
            SetBarcodeTimeoutValue = txtBarcodeTimeout.Text;
            SetLinkProfileValue = cmbLinkProfile.SelectedIndex.ToString();
            SetBaudRateValue = cmbOptionBaudRate.SelectedIndex.ToString();
            if (cbFrequencyAuto.IsChecked == true) SetFrequencyAutomaticValue = "1";
            else SetFrequencyAutomaticValue = "0";
            Task.Run(() =>
            {
                if (SetFrequencyAutomaticValue == "1")
                {
                    Device.SetFrequencyAutomatic();
                }
                else
                {
                    SetFrequencyAssigned();
                }
                Thread.Sleep(50);
                SetBuzzer();
                Thread.Sleep(50);
                SetIdleTime(SetIdleTimeValue);
                Thread.Sleep(50);
                SetSleepTime(SetSleepTimeValue);
                Thread.Sleep(50);
                SetAutoOffTime(SetAutoOffTimeValue);
                Thread.Sleep(50);
                SetQValue(SetQValueValue);
                Thread.Sleep(50);
                SetLinkProfile(SetLinkProfileValue);
                Thread.Sleep(50);
                SetBarcodeTimeout(SetBarcodeTimeoutValue);
                Thread.Sleep(50);
                SetBaudRate(SetBaudRateValue);
                Thread.Sleep(50);
            });
            if (cmbDefaultLinkProfile.SelectedIndex >= 0) Properties.Settings.Default.DefaultLinkProfile = cmbDefaultLinkProfile.SelectedIndex.ToString();
            if (int.TryParse(txtBatteryInterval.Text, out _)) Properties.Settings.Default.BatteryInterval = txtBatteryInterval.Text;
            Properties.Settings.Default.Save();
            btnOptionSave.IsEnabled = true;
            if (int.TryParse(txtBatteryInterval.Text, out SettingBatteryInterval) && SettingBatteryInterval > 0)
            {
                SetBatteryInterverTimer(SettingBatteryInterval);
                GetBatteryTimer.Start();
            }
            else if (GetBatteryTimer != null)
            {
                GetBatteryTimer.Stop();
            }
        }
        void InventoryTimerDo(object obj)
        {
            Device.StartBarcodeScan();
        }
        void EnabledInventoryItems(bool isEnabled)
        {
            cmbPowerGain.IsEnabled = isEnabled;
            cmbIvtSession.IsEnabled = isEnabled;
            cmbSessionFlag.IsEnabled = isEnabled;
            txtOperationTime.IsEnabled = isEnabled;
            cbDisplayPC.IsEnabled = isEnabled;
            cbContinuousMode.IsEnabled = isEnabled;
            cbReportRssi.IsEnabled = isEnabled;
            rbEPC.IsEnabled = isEnabled;
            rbTemperature.IsEnabled = isEnabled;
            rbMoisture.IsEnabled = isEnabled;
        }
        private void btnStartInventory_Click(object sender, RoutedEventArgs e)
        {
            if (btnInventory.Content.ToString() == "Start")
            {
                if (TagList != null && TagList.Count > 0)
                {
                    foreach (var item in TagList.Where(x => x.IsCheck == true))
                    {
                        item.IsCheck = false;
                    }
                }
                lblWarning.Visibility = Visibility.Hidden;
                STimerInventory.Start();
                InventoryTimer = new System.Threading.Timer(new System.Threading.TimerCallback(ShowTags), null, 0, 200);
                if (rbEPC.IsChecked == true)
                {
                    Device.StartInventory();
                }
                else if (rbTemperature.IsChecked == true)
                {
                    Device.StartInventoryTemperature();
                }
                else
                {
                    Device.StartInventoryMoisture();
                }
                EnabledInventoryItems(false);
                btnInventory.Content = "Stop";
                selectedTags.Clear();
                ckTagAll.IsChecked = false;
            }
            else
            {
                Device.Stop();
                STimerInventory.Stop();
                EnabledInventoryItems(true);
                btnInventory.Content = "Start";
            }
        }

        private void btnIvtClear_Click(object sender, RoutedEventArgs e)
        {
            lock (TagList)
            {
                TagList.Clear();
                dgTags.ItemsSource = null;
            }
            //dgTags.ItemsSource = null;
            //dgTags.Items.Clear();
            //dgTags.Items.Refresh();//多线程赋值必须
            readTagCount = 0;
            readTagCountBK = 0;
            lblTagCount.Content = "0";
            lblReadCount.Content = "0";
            lblTime.Content = "0";
            lblSpeed.Content = "0";
            inventorySecondCount = 0;
            txtLog.Text = "";
            countArr = new int[5];
            arrTop = -1;
        }
        UpdateFirmware updateFirmware;
        UpdateFirmware2 updateFirmware2;
        void UpdateFirmwareWinClosed(UpdateFirmwareType updateFirmwareType)
        {
            try
            {
                Dispatcher.Invoke((Action)(() =>
                {
                    if (updateFirmwareType == UpdateFirmwareType.UpdateFirmware) btnDisconnect_Click(null, null);
                    //if (isConnected)
                    //{
                    //    isConnected = false;
                    //    imgBattery.Source = new BitmapImage(new Uri("Resources/battery0.png", UriKind.RelativeOrAbsolute));
                    //    EnableControllers();
                    //    if (GetBatteryTimer != null) GetBatteryTimer.Stop();
                    //}
                }));
            }
            catch (Exception ex)
            {
                Console.WriteLine("UpdateFirmwareWinClosed: " + ex.Message);
            }
        }
        private void btnUpdateFirmware_Click(object sender, RoutedEventArgs e)
        {
            //UpdateFirmwareCommon(UpdateFirmwareType.UpdateFirmware, "Update Firmware");
            //if (GetTemperatureTimer.Enabled) GetTemperatureTimer.Stop();
            UpdateFirmwareCommon2(UpdateFirmwareType.UpdateFirmware, "Update Firmware");
        }
        private void btnRfidUpdateFirmware_Click(object sender, RoutedEventArgs e)
        {
            //UpdateFirmwareCommon(UpdateFirmwareType.RFIDUpdateFirmware, "RFID Update Firmware");
            //if (GetTemperatureTimer.Enabled) GetTemperatureTimer.Stop();
            UpdateFirmwareCommon2(UpdateFirmwareType.RFIDUpdateFirmware, "RFID Update Firmware");
        }
        void UpdateFirmwareCommon(UpdateFirmwareType updateFirmwareType, string title)
        {
            if (btnScanBarcode.Content.ToString() == "Stop Scan")
                MessageBox.Show("Please stop scan barcodes.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else if (btnInventory.Content.ToString() == "Stop")
                MessageBox.Show("Please stop inventory tags.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                updateFirmware = new UpdateFirmware(updateFirmwareType);
                updateFirmware.Owner = System.Windows.Application.Current.MainWindow;
                updateFirmware.Title = title;
                updateFirmware.winClose = UpdateFirmwareWinClosed;
                updateFirmware.ShowDialog();
            }
        }
        void UpdateFirmwareCommon2(UpdateFirmwareType updateFirmwareType, string title)
        {
            if (btnScanBarcode.Content.ToString() == "Stop Scan")
                MessageBox.Show("Please stop scan barcodes.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else if (btnInventory.Content.ToString() == "Stop")
                MessageBox.Show("Please stop inventory tags.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                updateFirmware2 = new UpdateFirmware2(updateFirmwareType);
                updateFirmware2.Owner = System.Windows.Application.Current.MainWindow;
                updateFirmware2.Title = title;
                updateFirmware2.winClose = UpdateFirmwareWinClosed;
                updateFirmware2.ShowDialog();
            }
        }
        string SetRssiValue;
        string SetSessionValue;
        string SetSessionFlagValue;
        private void cbReportRssi_Click(object sender, RoutedEventArgs e)
        {
            if (((CheckBox)sender).IsChecked == true)
            {
                SetRssiValue = ((int)Types.RssiOnOff.On).ToString();
                Device.SetRssiOnOff(Types.RssiOnOff.On);
            }
            else
            {
                SetRssiValue = ((int)Types.RssiOnOff.Off).ToString();
                Device.SetRssiOnOff(Types.RssiOnOff.Off);
            }
        }
        string SetContinuousModeValue;
        private void cbContinuousMode_Click(object sender, RoutedEventArgs e)
        {
            if (((CheckBox)sender).IsChecked == true)
            {
                SetContinuousModeValue = ((int)Types.InventoryMode.Continuous).ToString();
                Device.SetInventoryMode(Types.InventoryMode.Continuous);
            }
            else
            {
                SetContinuousModeValue = ((int)Types.InventoryMode.Single).ToString();
                Device.SetInventoryMode(Types.InventoryMode.Single);
            }
        }
        private void btnRead_Click(object sender, RoutedEventArgs e)
        {
            if (btnRead.Content.ToString() == "Read")
            {
                txtReadData.Text = txtReadMemory.Text = "";
                int readOffset, readLength;
                if (txtReadOffset.Text == "" || txtReadLength.Text == "" || !int.TryParse(txtReadOffset.Text.Trim(), out readOffset) || !int.TryParse(txtReadLength.Text.Trim(), out readLength) || readLength == 0)
                {
                    ShowMessage("Read parameters error.");
                    return;
                }
                btnRead.Content = "Stop";
                actionType = ActionType.Read;
                Device.SetOperationTime(txtReadOperationTime.Text);
                Thread.Sleep(50);
                Types.DataArea dataArea = rbReadEPC.IsChecked == true ? Types.DataArea.EPC :
                    rbReadTID.IsChecked == true ? Types.DataArea.TID :
                    rbReadUser.IsChecked == true ? Types.DataArea.User : Types.DataArea.Reserved;
                UInt32 ret = Device.Read(dataArea, readOffset, readLength, txtReadPwd.Text);
                if (ret != 0)
                {
                    ShowMessage("Read error");
                    btnRead.Content = "Read";
                }
            }
            else
            {
                btnRead.Content = "Read";
                Device.Stop();
            }
        }
        private void btnReadClear_Click(object sender, RoutedEventArgs e)
        {
            rbReadEPC.IsChecked = true;
            //cmbReadOffset.SelectedIndex = 2;
            //cmbReadLength.SelectedIndex = 2;
            txtReadOffset.Text = "2";
            txtReadLength.Text = "2";
            txtReadPwd.Text = "00000000";
            txtReadData.Text = "";
            lblReadRssi.Content = "";
            lblReadPhase.Content = "";
            txtReadMemory.Text = "";
        }
        private void btnWrite_Click(object sender, RoutedEventArgs e)
        {
            if (btnWrite.Content.ToString() == "Write")
            {
                int offset;
                if (txtWriteData.Text.Length == 0 || txtWriteData.Text.Length % 4 != 0)
                {
                    ShowMessage("Write Data must be entered in WORD unit.(1WORD = 16-bit)");
                }
                else if (txtWriteData.Text.Length / 4 > 32)
                {
                    ShowMessage("Write Data cannot be exceed 32WORD.(1WORD = 16-bit)");
                }
                else if (txtWriteOffset.Text == "")
                {
                    ShowMessage("Offset can not be null.");
                }
                else if (!int.TryParse(txtWriteOffset.Text, out offset))
                {
                    ShowMessage("Offset error.");
                }
                else
                {
                    actionType = ActionType.Write;
                    btnWrite.Content = "Stop";
                    Device.SetOperationTime(txtWriteOperationTime.Text);
                    Thread.Sleep(100);
                    //Device.SetPassword(txtWritePwd.Text);
                    Types.DataArea dataArea = rbWriteEPC.IsChecked == true ? Types.DataArea.EPC
                        : rbWriteTID.IsChecked == true ? Types.DataArea.TID
                        : rbWriteUser.IsChecked == true ? Types.DataArea.User : Types.DataArea.Reserved;
                    Thread.Sleep(50);
                    UInt32 ret = Device.Write(dataArea, offset, txtWriteData.Text, txtWritePwd.Text);
                    if (ret != 0)
                    {
                        ShowMessage("Write error");
                        btnWrite.Content = "Write";
                    }
                }
            }
            else
            {
                btnWrite.Content = "Write";
            }
        }
        private void btnWriteClear_Click(object sender, RoutedEventArgs e)
        {
            rbWriteEPC.IsChecked = true;
            cmbWriteOffset.SelectedIndex = 2;
            txtWriteData.Text = "";
            txtWritePwd.Text = "00000000";
            txtWriteReadData.Text = "";
            lblWriteRssi.Content = "";
            lblWritePhase.Content = "";
        }
        void ClearLockAndKillResult()
        {
            txtLockReadData.Text = "";
            lblLockRssi.Content = "";
            lblLockPhase.Content = "";
        }
        private void btnLock_Click(object sender, RoutedEventArgs e)
        {
            if (btnLock.Content.ToString() == "Stop")
            {
                btnLock.Content = "Lock";
            }
            else
            {
                ClearLockAndKillResult();
                if (txtLockPwd.Text.Trim() == "")
                {
                    ShowMessage("Password can not be null.");
                }
                else
                {
                    btnLock.Content = "Stop";
                    //Device.SetPassword(txtLockPwd.Text);
                    //Thread.Sleep(50);
                    Device.SetOperationTime(txtLockOperationTime.Text.Trim() == "" ? "0" : txtLockOperationTime.Text.Trim());
                    Thread.Sleep(50);
                    actionType = ActionType.Lock;
                    UInt32 ret = Device.Lock(cbLockKillPassword.IsChecked == true,
                        cbLockAccessPassword.IsChecked == true,
                        cbLockEPC.IsChecked == true,
                        cbLockTID.IsChecked == true,
                        cbLockUser.IsChecked == true,
                        txtLockPwd.Text);
                    if (ret != 0)
                    {
                        ShowMessage("Lock error");
                        btnLock.Content = "Lock";
                    }
                }
            }
        }
        private void btnUnLock_Click(object sender, RoutedEventArgs e)
        {
            if (btnUnLock.Content.ToString() == "Stop")
            {
                btnUnLock.Content = "Unlock";
            }
            else
            {
                ClearLockAndKillResult();
                if (txtLockPwd.Text.Trim() == "")
                {
                    ShowMessage("Password can not be null.");
                }
                else
                {
                    btnUnLock.Content = "Stop";
                    //Device.SetPassword(txtLockPwd.Text);
                    //Thread.Sleep(50);
                    Device.SetOperationTime(txtLockOperationTime.Text.Trim() == "" ? "0" : txtLockOperationTime.Text);
                    Thread.Sleep(50);
                    actionType = ActionType.Unlock;
                    UInt32 ret = Device.UnLock(cbLockKillPassword.IsChecked == true,
                        cbLockAccessPassword.IsChecked == true,
                        cbLockEPC.IsChecked == true,
                        cbLockTID.IsChecked == true,
                        cbLockUser.IsChecked == true,
                        txtLockPwd.Text);
                    if (ret != 0)
                    {
                        ShowMessage("Unlock error");
                        btnUnLock.Content = "Unlock";
                    }
                }
            }
        }
        private void btnPermaLock_Click(object sender, RoutedEventArgs e)
        {
            if (btnPermalock.Content.ToString() == "Stop")
            {
                btnPermalock.Content = "Permalock";
            }
            else
            {
                ClearLockAndKillResult();
                if (txtLockPwd.Text.Trim() == "")
                {
                    ShowMessage("Password can not be null.");
                }
                else
                {
                    btnPermalock.Content = "Stop";
                    //Device.SetPassword(txtLockPwd.Text);
                    //Thread.Sleep(100);
                    Device.SetOperationTime(txtLockOperationTime.Text.Trim() == "" ? "0" : txtLockOperationTime.Text);
                    Thread.Sleep(100);
                    actionType = ActionType.Permalock;
                    UInt32 ret = Device.PermaLock(cbLockKillPassword.IsChecked == true,
                        cbLockAccessPassword.IsChecked == true,
                        cbLockEPC.IsChecked == true,
                        cbLockTID.IsChecked == true,
                        cbLockUser.IsChecked == true,
                        txtLockPwd.Text);
                    if (ret != 0)
                    {
                        ShowMessage("Permalock error.");
                        btnPermalock.Content = "Permalock";
                    }
                }
            }
        }
        private void btnKill_Click(object sender, RoutedEventArgs e)
        {
            if (btnKill.Content.ToString() == "Stop")
            {
                btnKill.Content = "Kill";
            }
            else
            {
                ClearLockAndKillResult();
                if (txtLockPwd.Text.Trim() == "")
                {
                    ShowMessage("Password can not be null.");
                }
                else
                {
                    btnKill.Content = "Stop";
                    actionType = ActionType.Kill;
                    UInt32 ret = Device.Kill(txtLockPwd.Text);
                    if (ret != 0)
                    {
                        ShowMessage("Kill error.");
                        btnKill.Content = "Kill";
                    }
                }
            }
        }
        private void btnSetAccessPwd_Click(object sender, RoutedEventArgs e)
        {
            if (btnSetAccessPwd.Content.ToString() == "Stop")
            {
                btnSetAccessPwd.Content = "Set Access Password";
            }
            else
            {
                ClearLockAndKillResult();
                if (txtAccessPwd.Text.Trim() == "")
                {
                    ShowMessage("Access password can not be null.");
                }
                else
                {
                    actionType = ActionType.Set_Access_Password;
                    btnSetAccessPwd.Content = "Stop";
                    //Device.SetPassword(txtLockPwd.Text);//txtLockPwd可为空
                    //Thread.Sleep(50);
                    UInt32 ret = Device.SetAccessPassword(txtAccessPwd.Text, txtLockPwd.Text);
                    if (ret != 0)
                    {
                        ShowMessage("Set Access Password error.");
                        btnSetAccessPwd.Content = "Set Access Password";
                    }
                }
            }
        }
        private void btnSetKillPwd_Click(object sender, RoutedEventArgs e)
        {
            if (btnSetKillPwd.Content.ToString() == "Stop")
            {
                btnSetKillPwd.Content = "Set Kill Password";
            }
            else
            {
                ClearLockAndKillResult();
                if (txtKillPwd.Text.Trim() == "")
                {
                    ShowMessage("Kill Password can not be null.");
                }
                else
                {
                    actionType = ActionType.Set_Kill_Password;
                    btnSetKillPwd.Content = "Stop";
                    //Device.SetPassword(txtLockPwd.Text);//txtLockPwd可为空
                    //Thread.Sleep(50);
                    UInt32 ret = Device.SetKillPassword(txtKillPwd.Text, txtLockPwd.Text);
                    if (ret != 0)
                    {
                        ShowMessage("Set Kill Password error.");
                        btnSetKillPwd.Content = "Set Kill Password";
                    }
                }
            }
        }
        private void btnLockClear_Click(object sender, RoutedEventArgs e)
        {
            cbLockKillPassword.IsChecked = cbLockAccessPassword.IsChecked = cbLockEPC.IsChecked = cbLockTID.IsChecked = cbLockUser.IsChecked = false;
            txtLockPwd.Text = "00000000";
            txtAccessPwd.Text = "";
            txtKillPwd.Text = "";
            txtLockReadData.Text = "";
            lblLockRssi.Content = "";
            lblLockPhase.Content = "";
        }
        private void txtOperationTime_LostFocus(object sender, RoutedEventArgs e)
        {
            TextBox tb = (TextBox)sender;
            if (tb.Text == "") tb.Text = "0";
            SetOperationTime(tb.Text);
        }
        private void cmbPowerGain_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (cmbPowerGainStatus != DropDownStatus.Opened) return;
            ComboBox cmb = (ComboBox)sender;
            if (ReceivedPowerGainIndex != -1 && ReceivedPowerGainIndex == cmb.SelectedIndex) return;
            SetPowerGain(cmb.SelectedItem.ToString().Replace(" dB", ""));
            cmbPowerGainStatus = DropDownStatus.None;
        }
        private void cmbIvtSession_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            SetSession(cmbIvtSession.SelectedIndex.ToString());
        }
        private void cmbSessionFlag_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            SetSessionFlag(cmbSessionFlag.SelectedIndex.ToString());
        }

        private void cbDisplayPC_Checked(object sender, RoutedEventArgs e)
        {
            if (dgTags != null)
            {
                dgTags.Columns[3].Visibility = Visibility.Visible;
            }
        }

        private void cbDisplayPC_UnChecked(object sender, RoutedEventArgs e)
        {
            if (dgTags != null)
            {
                dgTags.Columns[3].Visibility = Visibility.Hidden;
            }
        }
        private void btnSearch_Click(object sender, RoutedEventArgs e)
        {
            if (ConnectBTDeviceTimer.Enabled) ConnectBTDeviceTimer.Stop();
            deviceSearchUI = new DeviceSearch();
            deviceSearchUI.Owner = System.Windows.Application.Current.MainWindow;
            deviceSearchUI.deviceSearchClose = DeviceSearchUIClose;
            deviceSearchUI.ShowDialog();
        }
        void DeviceSearchUIClose(SearchBluetoothDevice SearchBluetoothDevice)
        {
            currentBluetoothDevice = SearchBluetoothDevice.BluetoothDevice;
            lblBTName.Content = currentBluetoothDevice.DeviceName;
        }
        private void btnBTDisconnect_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (tbRight.SelectedIndex == 0 && btnInventory.Content.ToString() == "Stop") btnStartInventory_Click(null, null);
                if (tbRight.SelectedIndex == 1 && btnScanBarcode.Content.ToString() == "Stop Scan") btnScanBarcode_Click(null, null);
                if (connectStatus == ConnectStatus.Connected)
                {
                    _isBTDisConnClick = true;
                    Device.Disconnect();
                    lblWarning.Visibility = Visibility.Hidden;
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("btnBTDisconnect_Click: " + ex.Message);
            }
        }
        private void btnBTConnect_Click(object sender, RoutedEventArgs e)
        {
            ShowMessage("Connecting...");
            if (sender != null && ConnectBTDeviceTimer.Enabled)
            {
                ConnectBTDeviceTimer.Stop();
            }
            if (currentBluetoothDevice != null)
            {
                btnSearch.IsEnabled = btnBTConnect.IsEnabled = false;
                Device.Connect(currentBluetoothDevice);
            }
            else
            {
                ShowMessage("Please select a bluetooth device from the search page.");
            }
        }
        DropDownStatus cmbPowerGainStatus = DropDownStatus.None;

        private void cmbPowerGain_DropDownOpend(object sender, EventArgs e)
        {
            cmbPowerGainStatus = DropDownStatus.Opened;
        }

        private void tabTemperature_Clicked(object sender, MouseButtonEventArgs e)
        {
            Device.GetTemperature();
        }

        private void btnGetTemp_Click(object sender, RoutedEventArgs e)
        {
            Device.GetTemperature();
        }

        private void btnSetTemp_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                double threshold;
                if (double.TryParse(txtThreshold.Text, out threshold))
                {
                    Device.SetTemperature(cmbTemperatureOnOff.SelectedIndex, Convert.ToDouble(txtThreshold.Text), cmbOverrunAction.SelectedIndex);
                }
                else
                {
                    ShowMessage($"Set Temperature parameter threshold error.");
                }
            }
            catch (Exception ex)
            {
                ShowMessage($"Set Temperature failed.{ex.Message}");
            }
        }
        LEDStatus _LEDStatus = LEDStatus.None;
        private void btnLED_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (btnLED.Content.ToString() == "LED ON")
                {
                    var linq = TagList.Where(t => t.IsCheck == true).Select(t=>t.IsCheck).ToArray();
                    if (linq.Length > 0)
                    {
                        btnLED.Content = "LED OFF";
                        _LEDStatus = LEDStatus.On;
                        Task.Run(() =>
                        {
                            //ReadLEDByDataGridItems();
                            ReadLEDByTagList();
                        });
                    }
                }
                else
                {
                    _LEDStatus = LEDStatus.Off;
                    Task.Run(() =>
                    {
                        DateTime before = DateTime.Now;
                        Console.WriteLine(DateTime.Now.ToString("G") + "while() START");
                        while (true)
                        {
                            if (LedReading == false)
                            {
                                LedOff();
                                break;
                            }
                            else
                            {
                                TimeSpan diff = DateTime.Now.Subtract(before);
                                //Console.WriteLine("while() run time: " + diff.TotalSeconds.ToString());
                                if (diff.TotalSeconds >= 3)
                                {
                                    Console.WriteLine(DateTime.Now.ToString("G") + "while() end"+ diff.TotalSeconds.ToString());
                                    //3秒之后跳出循环
                                    LedOff();
                                    break;
                                }
                            }
                        }
                    });
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("btnLED_Click error:" + ex.Message);
            }
        }
        void LedOff()
        {
            Console.WriteLine(DateTime.Now.ToString("G") + "**LedOff**");
            this.Dispatcher.Invoke(() =>
            {
                btnLED.Content = "LED ON";
            });
            Thread.Sleep(200);
            Device.ClearEPCMaskData();
        }

        private void ReadLEDByDataGridItems()
        {
            string epc;
            string beforeEPCMaskData = "";
            while (_LEDStatus == LEDStatus.On)
            {
                for (int i = 0; i < selectedTags.Count; i++)
                {
                    if (_LEDStatus == LEDStatus.Off) break;
                    epc = selectedTags.ElementAt(i).Value.TagData.EPC;
                    if (beforeEPCMaskData != epc)
                    {
                        beforeEPCMaskData = epc;
                        Device.ClearEPCMaskData();
                        Thread.Sleep(200);
                        Device.SetEPCMaskData(32, epc, epc.Length * 4);
                        Thread.Sleep(200);
                        Console.WriteLine("Mask: " + epc);
                    }
                    Device.Read(Types.DataArea.User, 112, 1, "00000000");
                    Thread.Sleep(800);
                    Device.Stop();
                    Thread.Sleep(200);
                }
                Thread.Sleep(100);
            }
        }

        bool LedReading = false;
        private void ReadLEDByTagList()
        {
            string beforeEPCMaskData = "";
            while (_LEDStatus == LEDStatus.On)
            {
                for (int i = 0; i < TagList.Count; i++)
                {
                    if (_LEDStatus == LEDStatus.Off) break;
                    if (TagList[i].IsCheck == false) continue;
                    LedReading = true;
                    if (beforeEPCMaskData != TagList[i].TagData.EPC)
                    {
                        Console.WriteLine(DateTime.Now.ToString("G") + "...start...");
                        beforeEPCMaskData = TagList[i].TagData.EPC;
                        Device.ClearEPCMaskData();
                        //Console.WriteLine("ClearEPCMaskData");
                        Thread.Sleep(200);
                        Device.SetEPCMaskData(32, TagList[i].TagData.EPC, TagList[i].TagData.EPC.Length * 4);
                        //Console.WriteLine("SetEPCMaskData");
                        Thread.Sleep(200);
                        Console.WriteLine(DateTime.Now.ToString("G") + "Mask: " + TagList[i].TagData.EPC);
                    }
                    Device.Read(Types.DataArea.User, 112, 1, "00000000");
                    Console.WriteLine(DateTime.Now.ToString("G") + "Read");
                    Thread.Sleep(1000);
                    Device.Stop();
                    Console.WriteLine(DateTime.Now.ToString("G") + "...Stop...");
                    Thread.Sleep(200);
                    LedReading = false;
                }
                Thread.Sleep(100);
            }
        }

        private void ckTagAll_Click(object sender, RoutedEventArgs e)
        {
            CheckBox cb = (CheckBox)sender;
            for (int i = 0; i < TagList.Count; i++)
            {
                TagList[i].IsCheck = cb.IsChecked.Value;
            }
            dgTags.ItemsSource = null;
            dgTags.ItemsSource = TagList;
            //CheckBox cb = (CheckBox)sender;
            //DataGridTemplateColumn templeColumn = (DataGridTemplateColumn)dgTags.Columns[9];
            //for (int i = 0; i < dgTags.Items.Count; i++)
            //{
            //    FrameworkElement element = templeColumn.GetCellContent(dgTags.Items[i]);
            //    CheckBox ck = (CheckBox)templeColumn.CellTemplate.FindName("ckTag", element);
            //    if (ck != null)
            //    {
            //        ck.IsChecked = cb.IsChecked;
            //    }
            //}
        }
        private void ckTagAll_Checked(object sender, RoutedEventArgs e)
        {
            for (int i = 0; i < TagList.Count; i++)
            {
                TagList[i].IsCheck = true;
            }
            dgTags.ItemsSource = null;
            dgTags.ItemsSource = TagList;
            //CheckBox cb = (CheckBox)sender;
            //DataGridTemplateColumn templeColumn = (DataGridTemplateColumn)dgTags.Columns[9];
            //for (int i = 0; i < dgTags.Items.Count; i++)
            //{
            //    FrameworkElement element = templeColumn.GetCellContent(dgTags.Items[i]);
            //    CheckBox ck = (CheckBox)templeColumn.CellTemplate.FindName("ckTag", element);
            //    if (ck != null)
            //    {
            //        ck.IsChecked = cb.IsChecked;
            //    }
            //}
        }
        private void ckTagAll_UnChecked(object sender, RoutedEventArgs e)
        {
            for (int i = 0; i < TagList.Count; i++)
            {
                TagList[i].IsCheck = false;
            }
            dgTags.ItemsSource = null;
            dgTags.ItemsSource = TagList;
            //CheckBox cb = (CheckBox)sender;
            //DataGridTemplateColumn templeColumn = (DataGridTemplateColumn)dgTags.Columns[9];
            //for (int i = 0; i < dgTags.Items.Count; i++)
            //{
            //    FrameworkElement element = templeColumn.GetCellContent(dgTags.Items[i]);
            //    CheckBox ck = (CheckBox)templeColumn.CellTemplate.FindName("ckTag", element);
            //    if (ck != null)
            //    {
            //        ck.IsChecked = cb.IsChecked;
            //    }
            //}
        }
        Dictionary<string, Tag> selectedTags = new Dictionary<string, Tag>();
        private void ckTag_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                CheckBox cb = (CheckBox)sender;
                Tag t = (Tag)cb.DataContext;
                UpdateTagListByCheckBox(cb, t);
            }
            catch (Exception ex)
            {
                Console.WriteLine("ckTag_Checked: " + ex.Message);
            }
            //try
            //{
            //    CheckBox cb = (CheckBox)sender;
            //    Tag t = (Tag)cb.DataContext;
            //    string key = $"{t.TagData.PC}_{t.TagData.EPC}";
            //    if (!selectedTags.ContainsKey(key))
            //    {
            //        selectedTags.Add(key, t);
            //    }
            //    if (selectedTags.Count == dgTags.Items.Count)
            //    {
            //        ckTagAll.IsChecked = true;
            //    }
            //    Console.WriteLine($"ckTag_Checked: {selectedTags.Count}");
            //}
            //catch (Exception ex)
            //{
            //    Console.WriteLine("ckTag_Checked: " + ex.Message);
            //}
        }
        private void ckTag_Checked(object sender, RoutedEventArgs e)
        {
            try
            {
                CheckBox cb = (CheckBox)sender;
                Tag t = (Tag)cb.DataContext;
                UpdateTagListByCheckBox(cb, t);
            }
            catch (Exception ex)
            {
                Console.WriteLine("ckTag_Checked: " + ex.Message);
            }
            //try
            //{
            //    CheckBox cb = (CheckBox)sender;
            //    Tag t = (Tag)cb.DataContext;
            //    string key = $"{t.TagData.PC}_{t.TagData.EPC}";
            //    if (!selectedTags.ContainsKey(key))
            //    {
            //        selectedTags.Add(key, t);
            //    }
            //    if (selectedTags.Count == dgTags.Items.Count)
            //    {
            //        ckTagAll.IsChecked = true;
            //    }
            //    Console.WriteLine($"ckTag_Checked: {selectedTags.Count}");
            //}
            //catch (Exception ex)
            //{
            //    Console.WriteLine("ckTag_Checked: " + ex.Message);
            //}
        }

        private void ckTag_UnChecked(object sender, RoutedEventArgs e)
        {
            try
            {
                ckTagAll.IsChecked = false;
                CheckBox cb = (CheckBox)sender;
                Tag t = (Tag)cb.DataContext;
                UpdateTagListByCheckBox(cb, t);
            }
            catch (Exception ex)
            {
                Console.WriteLine("ckTag_UnChecked: " + ex.Message);
            }
            //CheckBox cb = (CheckBox)sender;
            //Tag t = (Tag)cb.DataContext;
            //string key = $"{t.TagData.PC}_{t.TagData.EPC}";
            //if (selectedTags.ContainsKey(key))
            //{
            //    selectedTags.Remove(key);
            //}
            //ckTagAll.IsChecked = selectedTags.Count == dgTags.Items.Count;
            //Console.WriteLine($"ckTag_UnChecked: {selectedTags.Count}");
        }
        private void UpdateTagListByCheckBox(CheckBox cb, Tag t)
        {
            int listCnt = TagList.Count;
            for (int i = 0; i < listCnt; i++)
            {
                if (TagList[i].TagData.EPC == t.TagData.EPC && TagList[i].TagData.PC == t.TagData.PC)
                {
                    TagList[i].IsCheck = cb.IsChecked.Value;
                    break;
                }
            }
            var linq = TagList.Where(d => d.IsCheck == true).ToArray();
            ckTagAll.IsChecked = linq.Length == listCnt;
        }


        private void cmbPorts_DropDownOpened(object sender, EventArgs e)
        {
            GetComPorts();
        }
        private void rbInventoryType_Checked(object sender, EventArgs e)
        {
            if (dgTags != null)
            {
                dgTags.Columns[1].Visibility = rbTemperature.IsChecked == true ? Visibility.Visible : Visibility.Hidden;
                dgTags.Columns[2].Visibility = rbMoisture.IsChecked == true ? Visibility.Visible : Visibility.Hidden;
                btnIvtClear_Click(null, null);
            }
        }
        private void txtWriteData_LostFocus(object sender, EventArgs e) { }
        private void txtHexData_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.Key == Key.Back) e.Handled = true;
            if (!(e.Key == Key.Back || (e.Key >= Key.D0 && e.Key <= Key.D9)))
            {
                e.Handled = true;
            }
        }
        private void txtNumberData_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.Key == Key.Back) e.Handled = true;
            if (!(e.Key == Key.Back || (e.Key >= Key.D0 && e.Key <= Key.D9) || (e.Key >= Key.NumPad0 && e.Key <= Key.NumPad9)))
            {
                e.Handled = true;
            }
        }
        private void txtHexData_PreviewTextInput(object sender, TextCompositionEventArgs e)
        {
            if (!Regex.IsMatch(e.Text, @"^[0-9A-Fa-f]+$"))
            {
                e.Handled = true;
            }
        }

        private void txtOperationTime_PreviewTextInput(object sender, TextCompositionEventArgs e)
        {
            try
            {
                if (!Regex.IsMatch(e.Text, @"^[0-9]+$"))
                {
                    e.Handled = true;
                }
                else
                {
                    int ret = 0;
                    TextBox tb = (TextBox)sender;
                    string text;
                    if (tb.SelectedText != "")
                    {
                        text = tb.Text.Remove(tb.SelectionStart, tb.SelectionLength).Insert(tb.SelectionStart, e.Text);
                    }
                    else
                    {
                        text = tb.Text.Insert(tb.CaretIndex, e.Text);
                    }
                    string[] range = tb.Tag.ToString().Split(',');
                    int min = int.Parse(range[0].Trim());
                    int max = int.Parse(range[1].Trim());
                    if (!(int.TryParse(text, out ret) && ret >= 0 && max >= min && ret <= max))
                    {
                        e.Handled = true;
                    }
                }
            }
            catch (Exception ex)
            {
                Log.Error("txtOperationTime_PreviewTextInput error: " + ex.Message);
            }
        }
        private void txtOperationTime_PreviewKeyDown(object sender, KeyEventArgs e)
        {
            TextBox tb = (TextBox)sender;
            if (e.Key == Key.Back ||
            e.Key == Key.Left ||
            e.Key == Key.Delete ||
            e.Key == Key.Right ||
            (tb.Name == "txtThreshold" && e.Key == Key.Decimal))
            {
                e.Handled = false;
            }
            else if ((e.Key >= Key.D0 && e.Key <= Key.D9 && e.KeyboardDevice.Modifiers != ModifierKeys.Shift) ||
                (e.Key >= Key.NumPad0 && e.Key <= Key.NumPad9))
            {
                e.Handled = false;
            }
            else
            {
                e.Handled = true;
            }
        }
        private void btnTagExport_Click(object sender, RoutedEventArgs e)
        {
            int count = TagList.Count;
            if (count == 0)
            {
                ShowMessage("There is no data to export!");
            }
            else if (count != Convert.ToInt32(lblTagCount.Content))
            {
                ShowMessage("Export data error!");
            }
            else
            {
                string fileName = (rbTemperature.IsChecked == true ? "Temperature_" : rbMoisture.IsChecked == true ? "Moisture" : "") +
                    "Tag_" + DateTime.Now.ToString("yyyyMMddHHmmssfff") + ".csv";
                Thread th = new Thread(ExportCSV);
                ExportParameters exportParameters = new ExportParameters()
                {
                    FileName = fileName,
                    DataGridObj = dgTags,
                    Encode = Encoding.Default
                };
                th.Start(exportParameters);
            }
        }
        private void btnExportBarcode_Click(object sender, RoutedEventArgs e)
        {
            if (dgBarcodes.Items.Count == 0)
            {
                ShowMessage("There is no data to export!");
            }
            else
            {
                string fileName = "Barcode_" + DateTime.Now.ToString("yyyyMMddHHmmssfff") + ".csv";
                Thread th = new Thread(ExportCSV);
                ExportParameters exportParameters = new ExportParameters()
                {
                    FileName = fileName,
                    DataGridObj = dgBarcodes,
                    Encode = rbASCII.IsChecked == true ? Encoding.ASCII : rbUTF8.IsChecked == true ? Encoding.UTF8 : Encoding.GetEncoding("Shift-JIS"),
                };
                th.Start(exportParameters);
            }
        }
        void ExportCSV(object obj)
        {
            try
            {
                ExportParameters exportParameters = (ExportParameters)obj;
                string savePath = Properties.Settings.Default.CsvDefaultPath;
                if (Directory.Exists(savePath) == false)
                {
                    MessageBox.Show(savePath + "folder does not exists.");
                }
                else
                {
                    SaveFileDialog sfd = new SaveFileDialog();
                    sfd.RestoreDirectory = true;
                    sfd.Filter = "csv(*.csv)|*.csv|all(*.*)|*.*";
                    sfd.FileName = exportParameters.FileName;
                    if (savePath != "") sfd.InitialDirectory = savePath;
                    //sfd.InitialDirectory = Application.StartupPath;
                    if (sfd.ShowDialog() == true)
                    {
                        FileStream fs = new FileStream(sfd.FileName, FileMode.Create);
                        //header
                        Dispatcher.Invoke((Action)(() =>
                        {
                            DataGrid dg = exportParameters.DataGridObj;
                            StreamWriter sw = new StreamWriter(fs, Encoding.Default);
                            //StreamWriter sw = new StreamWriter(fs, dg.Name == "dgTags" ? Encoding.Default : encoding);
                            string[] colsHeader = exportParameters.DataGridObj.Columns.Where(t => t.Visibility == Visibility.Visible && t.DisplayIndex != 9)
                                                            .Select(t => t.Header.ToString()).ToArray();
                            sw.Write(string.Join(",", colsHeader));
                            sw.Write(Environment.NewLine);
                            for (int i = 0; i < dg.Items.Count; i++)
                            {
                                for (int j = 0; j < dg.Columns.Count; j++)
                                {
                                    if (dg.Columns[j].Visibility == Visibility.Visible)
                                    {
                                        if (dg.Name == "dgTags")
                                        {
                                            switch (j)
                                            {
                                                case 0:
                                                    sw.Write($"\"{TagList[i].Index}\"");
                                                    break;
                                                case 1:
                                                    sw.Write($"\"{TagList[i].TagData.Temperature}\"");
                                                    break;
                                                case 2:
                                                    sw.Write($"\"{TagList[i].TagData.Moisture}\"");
                                                    break;
                                                case 3:
                                                    sw.Write($"\"{TagList[i].TagData.PC}\"");
                                                    break;
                                                case 4:
                                                    sw.Write($"\"{TagList[i].TagData.EPC}\"");
                                                    break;
                                                case 5:
                                                    sw.Write($"\"{TagList[i].TagData.Rssi.ToString("F1")}dB\"");
                                                    break;
                                                case 6:
                                                    sw.Write($"\"{TagList[i].TagData.Phase.ToString("F0")}°\"");
                                                    break;
                                                case 7:
                                                    sw.Write($"\"{TagList[i].TagData.Frequency}MHz\"");
                                                    break;
                                                case 8:
                                                    sw.Write($"\"{TagList[i].Count}\"");
                                                    break;
                                                default:
                                                    break;

                                            }
                                        }
                                        else
                                        {
                                            switch (j)
                                            {
                                                case 0:
                                                    sw.Write($"\"{BarcodeList[i].Index}\"");
                                                    break;
                                                case 1:
                                                    sw.Write($"\"{BarcodeList[i].BarcodeData}\"");
                                                    break;
                                                case 2:
                                                    sw.Write($"\"{BarcodeList[i].TypeName}\"");
                                                    break;
                                                case 3:
                                                    sw.Write($"\"{BarcodeList[i].Count}\"");
                                                    break;
                                                default:
                                                    break;

                                            }
                                            //TextBlock tb = (TextBlock)dg.Columns[j].GetCellContent(dg.Items[i]);
                                            //if (tb != null)
                                            //{
                                            //    sw.Write($"\"{tb.Text}\"");
                                            //}
                                        }
                                        if (j != dg.Columns.Count - 1) sw.Write(",");
                                    }
                                }
                                sw.Write(Environment.NewLine);
                            }
                            sw.Flush();
                            sw.Close();
                            ShowMessage("Export csv file succeed.");
                        }));
                    }
                }
            }
            catch (Exception ex)
            {
                ShowMessage("Export csv file failed." + ex.Message);
            }
        }
    }
    public class Tag
    {
        public int Index { get; set; }
        public TagData TagData { get; set; }
        public int Count { get; set; }
        public int TotalCount { get; set; }
        public bool IsCheck { get; set; } = false;

        public Tag(int index, TagData tagData, int count, int totalCount)
        {
            Index = index;
            TagData = tagData;
            Count = count;
            TotalCount = totalCount;
        }
    }

    public class Barcode
    {
        public int Index { get; set; }
        public string BarcodeData { get; set; }
        public string TypeName { get; set; }
        public int Count { get; set; }
    }
    public class Bind
    {
        public int Id { set; get; }
        public object Value { get; set; }
        public bool isChecked { get; set; }
    }
    public enum UpdateFirmwareType
    {
        UpdateFirmware,
        RFIDUpdateFirmware,
    }
    public enum ActionType
    {
        None,
        Read,
        Write,
        Lock,
        Unlock,
        Permalock,
        Kill,
        Set_Access_Password,
        Set_Kill_Password,
    }
    public class ExportParameters
    {
        public string FileName;
        public Encoding Encode;
        public DataGrid DataGridObj;
    }

    public enum ConnectTypes
    {
        None,
        COM,
        Bluetooth
    }
    public enum DropDownStatus
    {
        None,
        Opened
    }
    public enum ConnectStatus
    {
        Connecting,
        Connected,
        Disconnected,
    }
    public enum LEDStatus
    {
        None,
        On,
        Off,
    }
    public class SearchBluetoothDevice
    {
        public string DeviceIDShort { get; set; }
        public string DeviceName { get; set; }
        public BluetoothDevice BluetoothDevice { get; set; }
    }
}
