﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;
using AsReaderP252B;
using System.Configuration;
using System.Xml;
using System.Xml.Linq;
using System.Net;
using System.Threading;
using System.Text.RegularExpressions;

namespace AsReaderP252B_Demo
{
    /// <summary>
    /// UpdateFirmware.xaml 的交互逻辑
    /// </summary>
    public partial class UpdateFirmware : Window
    {
        string AckApp = "";
        string sURL = "";
        string xmlName = "";
        XmlDataProvider xmlDataProvider = new XmlDataProvider();
        string xPath = "";
        public delegate void WinClose(UpdateFirmwareType updateFirmwareType);
        public WinClose winClose = null;
        UpdateFirmwareType updateType;
        //线程通知
        private static AutoResetEvent resetEvent = new AutoResetEvent(false);
        public UpdateFirmware(UpdateFirmwareType updateFirmwareType)
        {
            InitializeComponent();
            SetDelegateMethod();
            updateType = updateFirmwareType;
            if (updateFirmwareType == UpdateFirmwareType.UpdateFirmware)
            {
                sURL = Properties.Settings.Default.UpdateFirmwareUrl; //ConfigurationManager.AppSettings["UpdateFirmwareUrl"];
                xmlName = Properties.Settings.Default.UpdateFirmwareXmlName; //ConfigurationManager.AppSettings["UpdateFirmwareXmlName"];
                xPath = Properties.Settings.Default.UpdateFirmwareXPath; //ConfigurationManager.AppSettings["UpdateFirmwareXPath"];
            }
            else
            {
                sURL = Properties.Settings.Default.RFIDUpdateFirmwareUrl;
                xmlName = Properties.Settings.Default.RFIDUpdateFirmwareXmlName;
                xPath = Properties.Settings.Default.RFIDUpdateFirmwareXPath;
            }
            xmlDataProvider.XPath = xPath;
            xmlDataProvider.IsAsynchronous = true;
            cmbBinFiles.DataContext = xmlDataProvider;
        }
        void GetFileList()
        {
            XmlDocument xmlDocument = new XmlDocument();
            xmlDocument.Load("https://camera.asreaderapps.com/mrxapp/W3/fw/W3.xml");
            XmlNodeList xmlNodeList = xmlDocument.SelectSingleNode("/W3_Fw").ChildNodes;
            Dictionary<string, string> dicXmlNodes = new Dictionary<string, string>();
            foreach (XmlNode node in xmlNodeList)
            {
                XmlElement nd = (XmlElement)node;
                string status = nd.GetAttribute("status");
                StringBuilder sb = new StringBuilder();
                sb.Append(status);
                sb.Append("_");
                int cnt = node.ChildNodes.Count;
                string text = "";
                string fileName = "";
                for (int i = 0; i < cnt; i++)
                {
                    XmlElement nd2 = (XmlElement)node.ChildNodes[i];
                    text = nd2.InnerText.Trim();
                    sb.Append(text);
                    if (nd2.Name == "") fileName = text;
                    if (i < cnt - 1) sb.Append("_");
                }
                dicXmlNodes.Add(fileName, sb.ToString());
            }
        }
        void SetDelegateMethod()
        {
            Device.whenReceivedData = ReceivedData;
        }
        #region delegate functions
        bool? SendUpdateSuccess = null;
        int SendPkgCnt = 0;
        void ReceivedData(byte[] buffer)
        {
            //请求固件更新地址
            //SUCCESS ACK APP1：~R<0000+8040000+\r\n
            //SUCCESS ACK APP2:  ~R<0000+8060000+\r\n
            //ERROR ACK:         ~R<0001+\r\n
            //(bin文件对半截)
            string rcvData = System.Text.Encoding.ASCII.GetString(buffer);
            if (rcvData.StartsWith("~R<0001"))
            {
                ReceivedRequestUpdateFirmwareResult("");
            }
            else if (rcvData.StartsWith("~R<0000"))
            {
                string ackApp = Regex.Match(rcvData, @"~R<0000(?<ACK>(\d*)?)\r\n", RegexOptions.Singleline).Groups["ACK"].Value;
                ReceivedRequestUpdateFirmwareResult(ackApp);
            }
            else if (buffer[1] == 0x57)
            {
                switch (buffer[2])
                {
                    case 0x3D:
                        //SUCCESS ACK：~W=0000+\r\n
                        //ERROR ACK:    ~W=0001+\r\n
                        ReceivedStartUpdateFirmwareResult(int.Parse(Encoding.ASCII.GetString(new byte[] { buffer[6] })));
                        break;
                    case 0x3E:
                        //SUCCESS ACK：~W > 0000 +\r\n
                        //ERROR   ACK: ~W > 0001 +\r\n
                        ReceivedUpdateFirmwareResult(int.Parse(Encoding.ASCII.GetString(new byte[] { buffer[6] })));
                        break;
                    case 0x3F:
                        //SUCCESS ACK：~W?0000+\r\n
                        //ERROR ACK: ~W ? 0001 +\r\n
                        ReceivedUpdateFirmwareCompleted(int.Parse(Encoding.ASCII.GetString(new byte[] { buffer[6] })));
                        break;
                }
            }
        }
        void ReceivedUpdateFirmwareResult(int resultCode)
        {
            Console.WriteLine($"Receive Update Firmware: {resultCode}");
            Dispatcher.Invoke((Action)delegate ()
            {
                if (resultCode == 0)
                {
                    resetEvent.Set();
                    SendPkgCnt++;
                    Thread thread = new Thread(UpdateProgressBar);
                    thread.Start(SendPkgCnt);
                    //if (SendPkgCnt == pbUpdateFirmware.Maximum) Device.UpdateFrimwareComplete();
                }
                else SendUpdateSuccess = false;
            });
        }
        void ReceivedRequestUpdateFirmwareResult(string ackApp)
        {
            Dispatcher.Invoke((Action)delegate ()
            {
                Console.WriteLine("Receive Request Update Firmware: " + ackApp);
                AckApp = ackApp;
                if (ackApp != "") Device.StartUpdateFirmware();
                else btnUpdate.IsEnabled = true;
            });
        }
        void ReceivedStartUpdateFirmwareResult(int resultCode)
        {
            Dispatcher.Invoke((Action)delegate ()
            {
                Console.WriteLine($"Receive Start Update Firmware: {resultCode}");
                if (resultCode == 0)
                {
                    System.Threading.Thread thread = new System.Threading.Thread(DoUpdateFirmware);
                    thread.Start();
                }
                else
                {
                    btnUpdate.IsEnabled = true;
                }
            });
        }
        void ReceivedUpdateFirmwareCompleted(int resultCode)
        {
            Dispatcher.Invoke((Action)delegate ()
            {
                MessageBox.Show(this.Title + (resultCode == 0 ? " succeed." : " failed."), "Tip", MessageBoxButton.OK, MessageBoxImage.Information);
                winClose?.Invoke(updateType);
                this.Close();
            });
        }
        void DoUpdateFirmware()
        {
            try
            {
                if (fileBuffer.Length == 0)
                {
                    Dispatcher.Invoke((Action)(() =>
                    {
                        MessageBox.Show("Bin file size error!", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                        btnUpdate.IsEnabled = true;
                    }));
                }
                else
                {
                    byte pkgNo = 0;
                    int pkgNoLength = 1;
                    int transferDataLength = 256;
                    int sumFileSize = 0;
                    int last = fileBuffer.Length / 2;
                    if (AckApp == "8040000")
                    {
                        //last half
                        sumFileSize = fileBuffer.Length / 2;
                        last = fileBuffer.Length;
                    }
                    byte[] transferFile;
                    while (sumFileSize < last)
                    {
                        int remainder = last - sumFileSize;
                        if (remainder < transferDataLength)
                        {
                            transferDataLength = remainder;
                        }
                        transferFile = new byte[transferDataLength];
                        //transferFile = new byte[pkgNoLength + transferDataLength];
                        //transferFile[0] = pkgNo;
                        Array.Copy(fileBuffer, sumFileSize, transferFile, 1, transferDataLength);
                        SendUpdateSuccess = null;
                        for (int i = 0; i < 5; i++)
                        {
                            if (Device.UpdateFirmware(pkgNo, transferFile) != 0)
                            {
                                Dispatcher.Invoke((Action)(() =>
                                {
                                    MessageBox.Show("Update error!", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                                    btnUpdate.IsEnabled = true;
                                }));
                                return;
                            }
                            resetEvent.WaitOne(2000);
                            if (SendUpdateSuccess.HasValue && SendUpdateSuccess.Value == true) break;
                            else if (i == 4 && (SendUpdateSuccess == null || SendUpdateSuccess.Value == false))
                            {
                                Dispatcher.Invoke((Action)(() =>
                                {
                                    MessageBox.Show(this.Title + " Failed.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                                    btnUpdate.IsEnabled = true;
                                }));
                                return;
                            }
                            else continue;
                        }
                        sumFileSize += transferDataLength;
                        if (pkgNo == 0xff) pkgNo = 0;
                        else pkgNo++;
                    }
                    Device.UpdateFirmwareComplete();
                }
            }
            catch (Exception ex)
            {
                btnUpdate.IsEnabled = true;
                Console.WriteLine("DoUpdateFirmware: " + ex.Message);
            }
        }
        void UpdateProgressBar(object val)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                int value = (int)val;
                int tmp = (int)(value * 100 / pbUpdateFirmware.Maximum);
                pbUpdateFirmware.Value = value;
                lblUpdateProgress.Content = tmp + "%";
            }));
        }
        #endregion delegate functions
        [Obsolete]
        private void btnGetFiles_Click(object sender, RoutedEventArgs e)
        {
            Button bt = (Button)sender;
            if (bt.Content.ToString() == "Cancel")
            {
                bt.Content = "Get Files";
            }
            else
            {
                bt.Content = "Cancel";
            }
            Thread thread = new Thread(GetFilesList);
            thread.Start();
        }
        void GetFilesList()
        {
            try
            {
                XmlDocument xmlDocument = new XmlDocument();
                xmlDocument.Load(sURL + "/" + xmlName);
                xmlDataProvider.Document = xmlDocument;
                xmlDataProvider.XPath = xPath;
                MessageBox.Show("Get bin files succeed.", "Tips", MessageBoxButton.OK, MessageBoxImage.Information);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                MessageBox.Show("Get bin files failed.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            }
            Dispatcher.Invoke((Action)(() =>
            {
                btnGetFiles.Content = "Get Files";
            }));
        }
        byte[] fileBuffer = { };
        private void btnUpdate_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                btnUpdate.IsEnabled = false;
                if (cmbBinFiles.SelectedIndex < 0)
                {
                    MessageBox.Show("Please select a file for update.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                    btnUpdate.IsEnabled = true;
                }
                else
                {
                    fileBuffer = DownloadBinFile(cmbBinFiles.SelectedValue.ToString());
                    if (fileBuffer == null || fileBuffer.Length == 0)
                    {
                        MessageBox.Show("Bin file not exists.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                        btnUpdate.IsEnabled = true;
                    }
                    else
                    {
                        pbUpdateFirmware.Value = 0;
                        pbUpdateFirmware.Maximum = (int)Math.Ceiling(fileBuffer.Length / 2 / 256.0);
                        UInt32 ret = Device.RequestUpdateFirmware();
                        if (ret != 0)
                        {
                            MessageBox.Show("Request update frimware error.", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                            btnUpdate.IsEnabled = true;
                        }
                    }
                }

            }
            catch (Exception ex)
            {
                btnUpdate.IsEnabled = true;
            }
        }
        private byte[] DownloadBinFile(string fileName)
        {
            if (fileName == "") return null;
            try
            {
                WebClient webClient = new WebClient();
                return webClient.DownloadData(sURL + "/" + fileName.Trim());
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                MessageBox.Show("Download " + fileName + " failed.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                throw ex;
            }
        }
    }
}
