﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;
using System.Net;
using System.Threading;
using System.IO;
using System.Xml;
using System.Text.RegularExpressions;

namespace AsReaderP252B_Demo
{
    /// <summary>
    /// UpdateFirmware2.xaml 的交互逻辑
    /// </summary>
    public partial class UpdateFirmware2 : Window
    {
        string AckApp = "";
        string sURL = "";
        string xmlName = "";
        string xPath = "";
        public delegate void WinClose(UpdateFirmwareType updateFirmwareType);
        public WinClose winClose = null;
        string binPath = AppDomain.CurrentDomain.BaseDirectory;
        XmlDataProvider xmlDataProvider = new XmlDataProvider();
        public delegate void ProgressBarSetter(ProgressBar pb, Label lbl, double value);
        public UpdateFirmwareType UpdateType;
        //线程通知
        private static AutoResetEvent resetEvent = new AutoResetEvent(false);
        public UpdateFirmware2()
        {
            InitializeComponent();
            SetDelegateMethod();
            sURL = Properties.Settings.Default.UpdateFirmwareUrl;
            xmlName = Properties.Settings.Default.UpdateFirmwareXmlName;
            xPath = Properties.Settings.Default.UpdateFirmwareXPath;
            xmlDataProvider.XPath = xPath;
            xmlDataProvider.IsAsynchronous = true;
            dgBinFiles.DataContext = xmlDataProvider;
            btnGetFiles.RaiseEvent(new RoutedEventArgs(Button.ClickEvent));
        }
        public UpdateFirmware2(UpdateFirmwareType updateFirmwareType)
        {
            InitializeComponent();
            SetDelegateMethod();
            UpdateType = updateFirmwareType;
            if (updateFirmwareType == UpdateFirmwareType.UpdateFirmware)
            {
                sURL = Properties.Settings.Default.UpdateFirmwareUrl;
                xmlName = Properties.Settings.Default.UpdateFirmwareXmlName;
                xPath = Properties.Settings.Default.UpdateFirmwareXPath;
            }
            else
            {
                sURL = Properties.Settings.Default.RFIDUpdateFirmwareUrl;
                xmlName = Properties.Settings.Default.RFIDUpdateFirmwareXmlName;
                xPath = Properties.Settings.Default.RFIDUpdateFirmwareXPath;
            }
            xmlDataProvider.XPath = xPath;
            xmlDataProvider.IsAsynchronous = true;
            dgBinFiles.DataContext = xmlDataProvider;
            btnGetFiles.RaiseEvent(new RoutedEventArgs(Button.ClickEvent));
        }
        void SetDelegateMethod()
        {
            Device.whenReceivedData = ReceivedData;
        }
        private void ckFile_click(object sender, RoutedEventArgs e)
        {
            try
            {
                CheckBox sourceCk = (CheckBox)sender;
                if (sourceCk.IsChecked == false) return;
                DataGridTemplateColumn templeColumn = (DataGridTemplateColumn)dgBinFiles.Columns[0];
                if (templeColumn == null) return;
                for (int i = 0; i < dgBinFiles.Items.Count; i++)
                {
                    FrameworkElement element = templeColumn.GetCellContent(dgBinFiles.Items[i]);
                    CheckBox chk = (CheckBox)templeColumn.CellTemplate.FindName("ckFile", element);
                    if (chk != null)
                    {
                        if (sourceCk == chk) continue;
                        chk.IsChecked = false;
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("ckFile_click" + ex.Message);
            }
        }
        byte[] ReadFileToBuffers(string file)
        {
            try
            {
                if (File.Exists(file)) return File.ReadAllBytes(file);
                else return null;
            }
            catch (Exception ex)
            {
                Console.WriteLine("ReadFileToBuffers: " + ex.Message);
                return null;
            }
        }
        string GetUpdateFileName()
        {
            string fileName = "";
            try
            {
                DataGridTemplateColumn templeColumn = (DataGridTemplateColumn)dgBinFiles.Columns[0];
                if (templeColumn != null)
                {
                    for (int i = 0; i < dgBinFiles.Items.Count; i++)
                    {
                        FrameworkElement element = templeColumn.GetCellContent(dgBinFiles.Items[i]);
                        CheckBox chk = (CheckBox)templeColumn.CellTemplate.FindName("ckFile", element);
                        if (chk.IsChecked == true && chk.Tag != null)
                        {
                            fileName = ((XmlElement)chk.Tag).InnerText.Trim();
                            break;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("GetUpdateFileName: " + ex.Message);
            }
            return fileName;
        }
        byte[] fileBuffer = { };
        private void btnUpdate_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                btnUpdate.IsEnabled = false;
                string file = GetUpdateFileName();
                if (file == "")
                {
                    MessageBox.Show("Please select a file to update.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                    btnUpdate.IsEnabled = true;
                }
                else if (!File.Exists(binPath + file))
                {
                    MessageBox.Show($"{file} file does not exist.Please download it.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                    btnUpdate.IsEnabled = true;
                }
                else
                {
                    fileBuffer = ReadFileToBuffers(binPath + file);
                    if (fileBuffer == null || fileBuffer.Length == 0)
                    {
                        MessageBox.Show($"{file} file is invalid.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                        btnUpdate.IsEnabled = true;
                    }
                    else
                    {
                        pbUpdateFirmware.Value = 0;
                        UInt32 ret = 0;
                        if (UpdateType == UpdateFirmwareType.UpdateFirmware)
                        {
                            pbUpdateFirmware.Maximum = (int)Math.Ceiling(fileBuffer.Length / 2 / 256.0);
                            ret = Device.RequestUpdateFirmware();
                        }
                        else
                        {
                            pbUpdateFirmware.Maximum = (int)Math.Ceiling(fileBuffer.Length / 256.0);
                            ret = Device.RequestUpdateRFIDFirmware();
                        }
                        if (ret != 0)
                        {
                            MessageBox.Show("Request update frimware error.", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                            btnUpdate.IsEnabled = true;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("btnUpdate_Click: " + ex.Message);
                btnUpdate.IsEnabled = true;
            }
        }
        private void btnDownload_click(object sender, RoutedEventArgs e)
        {
            Button button = (Button)sender;
            try
            {
                button.IsEnabled = false;
                DataGrid datagrid = (DataGrid)(((Button)sender).CommandParameter);
                var item = button.Tag;
                DataGridTemplateColumn templeColumn = (DataGridTemplateColumn)datagrid.Columns[4];
                if (templeColumn == null) return;
                FrameworkElement element = templeColumn.GetCellContent(item);
                ProgressBar pb = (ProgressBar)templeColumn.CellTemplate.FindName("pbDown", element);
                pb.Value = 0;
                Label lblProgress = (Label)templeColumn.CellTemplate.FindName("lblProgress", element);
                lblProgress.Content = "0 %";
                string fileName = ((TextBlock)datagrid.Columns[3].GetCellContent(item)).Text.Trim();
                //if (!(File.Exists(binPath + fileName) && !isDownloadAll &&
                //    MessageBox.Show($"{fileName} file already exist.Do you want to download it again?", "Tips", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.No))
                //{
                //    Download(pb, lblProgress, fileName);
                //}
                //if(isDownloadAll) Download(pb, lblProgress, fileName);
                Download(pb, lblProgress, fileName, button);
            }
            catch (Exception ex)
            {
                button.IsEnabled = true;
                Console.WriteLine("btnDownload_click: " + ex.Message);
            }
        }
        void Download2(ProgressBar pbDown, Label lbl, string fileName, Button button)
        {
            try
            {
                //ThreadPool.SetMaxThreads(1, 1);
                //ThreadPool.QueueUserWorkItem(obj =>
                //{
                WebRequest request = WebRequest.Create(sURL + "/" + fileName);
                WebResponse respone = request.GetResponse();
                Dispatcher.Invoke((Action)(() => { pbDown.Maximum = respone.ContentLength; }));
                Stream netStream = respone.GetResponseStream();
                Stream fileStream = new FileStream(binPath + fileName, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite);
                //var stream = new FileStream(binPath + fileName, FileMode.Open, FileAccess.Read, FileShare.Write)
                byte[] read = new byte[1024];
                long progressBarValue = 0;
                int realReadLen = netStream.Read(read, 0, read.Length);
                while (realReadLen > 0)
                {
                    fileStream.Write(read, 0, realReadLen);
                    progressBarValue += realReadLen;
                    SetProgressBar(pbDown, lbl, progressBarValue, button);
                    realReadLen = netStream.Read(read, 0, read.Length);
                }
                netStream.Close();
                fileStream.Close();
                respone.Dispose();
                // });
            }
            catch (WebException webEx)
            {
                button.IsEnabled = true;
                MessageBox.Show("Failed to download file due to network exception.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            }
            catch (IOException ioEx)
            {
                button.IsEnabled = true;
                Console.WriteLine("Download: " + ioEx.Message);
            }
            catch (Exception ex)
            {
                button.IsEnabled = true;
                MessageBox.Show($"Download {fileName} failed.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                if (File.Exists(binPath + fileName)) File.Delete(binPath + fileName);
                Console.WriteLine("Download: " + ex.Message);
            }
            isDownloadAll = false;
        }
        void Download(ProgressBar pbDown, Label lbl, string fileName, Button button)
        {
            ThreadPool.QueueUserWorkItem(obj =>
            {
                try
                {
                    System.GC.Collect();
                    HttpWebRequest request = (HttpWebRequest)WebRequest.Create(sURL + "/" + fileName);
                    //request.KeepAlive = false;
                    HttpWebResponse response = (HttpWebResponse)request.GetResponse();
                    //request.ReadWriteTimeout = 60000;
                    Dispatcher.Invoke((Action)(() => { pbDown.Maximum = response.ContentLength; }));
                    Stream fileStream = new FileStream(binPath + fileName, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite);

                    Stream netStream = response.GetResponseStream();
                    byte[] read = new byte[1024];
                    long progressBarValue = 0;
                    int realReadLen = netStream.Read(read, 0, read.Length);
                    while (realReadLen > 0)
                    {
                        fileStream.Write(read, 0, realReadLen);
                        progressBarValue += realReadLen;
                        SetProgressBar(pbDown, lbl, progressBarValue, button);
                        realReadLen = netStream.Read(read, 0, read.Length);
                    }
                    netStream.Close();
                    fileStream.Close();
                    response.Dispose();
                    request.Abort();
                }
                catch (WebException webEx)
                {
                    Dispatcher.Invoke((Action)(() => { button.IsEnabled = true; }));
                    MessageBox.Show("Failed to download file due to network exception.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                }
                catch (IOException ioEx)
                {
                    Dispatcher.Invoke((Action)(() => { button.IsEnabled = true; }));
                    Console.WriteLine("Download: " + ioEx.Message);
                }
                catch (Exception ex)
                {
                    Dispatcher.Invoke((Action)(() => { button.IsEnabled = true; }));
                    MessageBox.Show($"Download {fileName} failed.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                    if (File.Exists(binPath + fileName)) File.Delete(binPath + fileName);
                    Console.WriteLine("Download: " + ex.Message);
                }
                isDownloadAll = false;
            });
        }
        public void SetProgressBar(ProgressBar pbProgress, Label lblProgress, double value, Button button)
        {
            Dispatcher.Invoke((Action)(() =>
            {
                pbProgress.Value = value;
                int val = (int)value;
                int tmp = (int)(val * 100 / pbProgress.Maximum);
                lblProgress.Content = tmp + " %";
                if (tmp == 100 && button != null) button.IsEnabled = true;
            }));
        }
        private void btnGetFiles_Click(object sender, RoutedEventArgs e)
        {
            if (btnGetFiles.Content.ToString() == "Cancel")
            {
                btnGetFiles.Content = "Get Files";
            }
            else {
                btnGetFiles.Content = "Cancel";
                Thread thread = new Thread(TGetFilesList);
                thread.Start();
            }
        }
        bool isDownloadAll = false;
        private void btnDownloadAll_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                int processorCount = System.Environment.ProcessorCount;
                bool ret = ThreadPool.SetMaxThreads(processorCount, processorCount);
                if (!ret) Console.WriteLine("ProcessorCount: " + System.Environment.ProcessorCount);
                DataGridTemplateColumn templeColumn = (DataGridTemplateColumn)dgBinFiles.Columns[5];
                if (templeColumn != null)
                {
                    for (int i = 0; i < dgBinFiles.Items.Count; i++)
                    {
                        FrameworkElement element = templeColumn.GetCellContent(dgBinFiles.Items[i]);
                        Button btnDl = (Button)templeColumn.CellTemplate.FindName("btnDownload", element);
                        if (btnDl != null)
                        {
                            isDownloadAll = true;
                            btnDl.RaiseEvent(new RoutedEventArgs(Button.ClickEvent));
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("btnDownloadAll_Click: " + ex.Message);
            }
        }
        void TGetFilesList()
        {
            try
            {
                XmlDocument xmlDocument = new XmlDocument();
                xmlDocument.Load(sURL + "/" + xmlName);
                xmlDataProvider.Document = xmlDocument;
                //xmlDataProvider.XPath = xPath;
            }
            catch (WebException webEx)
            {
                Dispatcher.Invoke((Action)(() =>
                {
                    MessageBox.Show("Failed to get file list due to network exception.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                }));
            }
            catch (Exception ex)
            {
                Dispatcher.Invoke((Action)(() =>
                {
                    MessageBox.Show("Get file failed.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                }));
                Console.WriteLine(ex.Message);
            }
            Dispatcher.Invoke((Action)(() =>
            {
                btnGetFiles.Content = "Get Files";
            }));
        }
        bool? SendUpdateSuccess = null;
        int SendPkgCnt = 0;
        void ReceivedData(byte[] buffer)
        {
            //请求固件更新地址
            //SUCCESS ACK APP1：~R<0000+8040000+\r\n
            //SUCCESS ACK APP2:  ~R<0000+8060000+\r\n
            //ERROR ACK:         ~R<0001+\r\n
            //(bin文件对半截)
            string rcvData = System.Text.Encoding.ASCII.GetString(buffer);
            if (rcvData.StartsWith("~R<0001"))
            {
                ReceivedRequestUpdateFirmwareResult("");
            }
            else if (rcvData.StartsWith("~R<0000"))
            {
                string ackApp = Regex.Match(rcvData, @"~R<0000(?<ACK>(\d*)?)\r\n", RegexOptions.Singleline).Groups["ACK"].Value;
                ReceivedRequestUpdateFirmwareResult(ackApp);
            }
            else if (buffer[1] == 0x57)
            {
                switch (buffer[2])
                {
                    case 0x3D:
                        //SUCCESS ACK：~W=0000+\r\n
                        //ERROR ACK:    ~W=0001+\r\n
                        ReceivedStartUpdateFirmwareResult(int.Parse(Encoding.ASCII.GetString(new byte[] { buffer[6] })));
                        break;
                    case 0x3E:
                        //SUCCESS ACK：~W > 0000 +\r\n
                        //ERROR   ACK: ~W > 0001 +\r\n
                        ReceivedUpdateFirmwareResult(int.Parse(Encoding.ASCII.GetString(new byte[] { buffer[6] })));
                        break;
                    case 0x7C:
                        //SUCCESS ACK：~W|0000+\r\n
                        //ERROR ACK: ~W|0001+\r\n
                    case 0x3F:
                        //SUCCESS ACK：~W?0000+\r\n
                        //ERROR ACK: ~W ? 0001 +\r\n
                        ReceivedUpdateFirmwareCompleted(int.Parse(Encoding.ASCII.GetString(new byte[] { buffer[6] })));
                        break;
                    case 0x7B:
                        //SUCCESS ACK：~W{0000+\r\n
                        //ERROR ACK: ~W{0001+\r\n
                        ReceivedRequestUpdateRFIDFirmwareResult(int.Parse(Encoding.ASCII.GetString(new byte[] { buffer[6] })));
                        break;
                    case 0x5E:
                        //SUCCESS ACK：~ w^0000+\r\n
                        //ERROR ACK: ~ w^0001+\r\n
                        ReceivedUpdateFirmwareResult(int.Parse(Encoding.ASCII.GetString(new byte[] { buffer[6] })));
                        break;
                }
            }
        }
        void ReceivedUpdateFirmwareResult(int resultCode)
        {
            Console.WriteLine($"Receive Update Firmware: {resultCode}");
            Dispatcher.Invoke((Action)delegate ()
            {
                if (resultCode == 0)
                {
                    SendUpdateSuccess = true;
                    SendPkgCnt++;
                    SetProgressBar(pbUpdateFirmware, lblUpgradeProgress, SendPkgCnt, null);
                    //Thread thread = new Thread(UpdateProgressBar);
                    //thread.Start(SendPkgCnt);
                    //if (SendPkgCnt == pbUpdateFirmware.Maximum) Device.UpdateFrimwareComplete();
                }
                else SendUpdateSuccess = false;
                resetEvent.Set();
            });
        }
        void ReceivedRequestUpdateFirmwareResult(string ackApp)
        {
            Dispatcher.Invoke((Action)delegate ()
            {
                Console.WriteLine("Receive Request Update Firmware: " + ackApp);
                AckApp = ackApp;
                if (ackApp != "") Device.StartUpdateFirmware();
                else btnUpdate.IsEnabled = true;
            });
        }
        void ReceivedRequestUpdateRFIDFirmwareResult(int resultCode)
        {
            Dispatcher.Invoke((Action)delegate ()
            {
                Console.WriteLine("Receive Request Update RFID Firmware result Code: " + resultCode.ToString());
                if (isUpdating) return;
                if (resultCode != 0) btnUpdate.IsEnabled = true;
                else
                {
                    System.Threading.Thread thread = new System.Threading.Thread(DoUpdateFirmware);
                    thread.Start();
                }
            });
        }
        void ReceivedStartUpdateFirmwareResult(int resultCode)
        {
            Dispatcher.Invoke((Action)delegate ()
            {
                Console.WriteLine($"Receive Start Update Firmware: {resultCode}");
                if (resultCode == 0)
                {
                    System.Threading.Thread thread = new System.Threading.Thread(DoUpdateFirmware);
                    thread.Start();
                }
                else
                {
                    btnUpdate.IsEnabled = true;
                }
            });
        }
        void ReceivedUpdateFirmwareCompleted(int resultCode)
        {
            Dispatcher.Invoke((Action)delegate ()
            {
                MessageBox.Show(this.Title + (resultCode == 0 ? " succeed." : " failed."), "Tip", MessageBoxButton.OK, MessageBoxImage.Information);
                if (resultCode == 0)
                {
                    if (UpdateType == UpdateFirmwareType.RFIDUpdateFirmware) Device.RestartRFIDFirmware();
                    winClose?.Invoke(UpdateType);
                    this.Close();
                }
            });
        }
        bool isUpdating = false;
        void DoUpdateFirmware()
        {
            try
            {
                isUpdating = true;
                if (fileBuffer.Length == 0)
                {
                    Dispatcher.Invoke((Action)(() =>
                    {
                        MessageBox.Show("Bin file size error!", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                        btnUpdate.IsEnabled = true;
                    }));
                }
                else
                {
                    uint pkgNo = 1;
                    //int pkgNoLength = 1;
                    int transferDataLength = 256;
                    int sumFileSize = 0;
                    int last = fileBuffer.Length;
                    if (UpdateType == UpdateFirmwareType.UpdateFirmware)
                    {
                        last = fileBuffer.Length / 2;
                        pkgNo = 0;
                    }
                    if (AckApp == "8040000" && UpdateType == UpdateFirmwareType.UpdateFirmware)
                    {
                        //last half
                        sumFileSize = fileBuffer.Length / 2;
                        last = fileBuffer.Length;
                    }
                    byte[] transferFile;
                    while (sumFileSize < last)
                    {
                        int remainder = last - sumFileSize;
                        if (remainder < transferDataLength)
                        {
                            transferDataLength = remainder;
                        }
                        transferFile = new byte[transferDataLength];
                        //transferFile = new byte[pkgNoLength + transferDataLength];
                        //transferFile[0] = pkgNo;
                        Array.Copy(fileBuffer, sumFileSize, transferFile, 0, transferDataLength);
                        SendUpdateSuccess = null;
                        uint ret = 0;
                        for (int i = 0; i < 5; i++)
                        {
                            if (UpdateType == UpdateFirmwareType.UpdateFirmware) ret = Device.UpdateFirmware((byte)pkgNo, transferFile);
                            else ret = Device.UpdateRFIDFirmware(pkgNo, transferFile);
                            if (ret != 0)
                            {
                                Dispatcher.Invoke((Action)(() =>
                                {
                                    MessageBox.Show("Update error!", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                                    btnUpdate.IsEnabled = true;
                                }));
                                isUpdating = false;
                                return;
                            }
                            resetEvent.WaitOne(2000);
                            if (SendUpdateSuccess.HasValue && SendUpdateSuccess.Value == true) break;
                            else if (i == 4 && (SendUpdateSuccess == null || SendUpdateSuccess.Value == false))
                            {
                                Dispatcher.Invoke((Action)(() =>
                                {
                                    MessageBox.Show(this.Title + " Failed.", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                                    btnUpdate.IsEnabled = true;
                                }));
                                isUpdating = false;
                                return;
                            }
                            else continue;
                        }
                        sumFileSize += transferDataLength;
                        if (pkgNo == 0xff && UpdateType == UpdateFirmwareType.UpdateFirmware)
                        {
                            pkgNo = 0;
                        }
                        else pkgNo++;
                    }
                    if (UpdateType == UpdateFirmwareType.UpdateFirmware) Device.UpdateFirmwareComplete();
                    else Device.UpdateRFIDFirmwareComplete();
                }
            }
            catch (Exception ex)
            {
                btnUpdate.IsEnabled = true;
                Console.WriteLine("DoUpdateFirmware: " + ex.Message);
            }
            isUpdating = false;
        }
    }
}
